/*--------------------------------------------------------------*/
/* defs.h							*/
/*	Definitions of structures, etc., used in the router	*/
/*	files.							*/
/*								*/
/* Copyright 2001 - Jennifer Roberts and Tim Edwards		*/	
/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/
/* Global Definitions						*/
/*--------------------------------------------------------------*/

/* The following values can be manipulated to reflect the size 	*/
/* of the cells used in the design.				*/

#define PENALTY		20 // cost per channel above target minimum
#define ALIGN_PENALTY	5  // cost per lambda of row misalignment
#define CHANNELS	1  // target minimum number of channels
#define PORT_WIDTH	3  // height/width of a wire/port
#define PORT_SPACE	4  // minimum spacing of a wire/port
#define VIA_W		4  // width of a via
#define VIA_HT		4  // height of a via
#define BUFFER		8  // distance channel vias can intrude into boxes
#define SPACER_WIDTH	8  // width of spacer cells in non-fet_designs
#define CONTACT_SIZE	4  // width,height for a substrate/well contact

/* Other adaptable constants					*/

#define PSTEP		1000
#define SPACERS		0		/* 3 */
#define RSLACK		60		/* 4 */

/* For GA version */

#define POPULATION	256
#define MUTATION_RATE	10		/* in percent */

/*------------------------------------------------------------------*/

// distance btw top of row 1 boxes and bottom of the lowest channel:
#define BOX_TO_CHAN	(m1_space - (fet_design ? 1 : 0) + (via_width - m1_width) / 2)

/* PORT_SEP = min distance between x-coords of adjacent ports.		*/
/* CHAN_SEP = min distance between bottoms of adjacent channels.	*/
#define PORT_SEP	(via_width + m2_space)
#define CHAN_SEP	(via_height + m1_space - V_ADJUST)

// coordinates needed to center via around wire
#define VIA_X(x)	(x - (via_width - m2_width) / 2)
#define VIA_Y(y)	(y - (via_height - m1_width) / 2)

#define V_ADJUST	((via_height - m1_width) % 2)
#define H_ADJUST	((via_width - m2_width) % 2)

#define ROWS		2

#define FETPORTS	4	// drain, gate, source, bulk

#define IDRAIN		0	// order of FET ports in SPICE format
#define IGATE		1
#define ISOURCE		2
#define IBULK		3

#define GND		0
#define VDD		1

#define IF_ON_TOP	2
#define IF_ON_BOTTOM	3

#define LOWER		0
#define UPPER		1

#define MAX_BOX_WIDTH	1000
#define MAX_BOX_HEIGHT	1000

#define PTP_TO_RIGHT	-1
#define PTP_TO_LEFT	-2

// from contraints.c

#define LEFT		1
#define RIGHT		2
#define TOP		3
#define BOTTOM		4
#define TOGETHER	5

// Types of metal used for routing

#define METAL1		0
#define METAL2		1

// MIN and MAX and ABS
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) < (b)) ? (b) : (a))
#define ABS(a)   (((a) < 0) ? -(a) : (a))

// Classes of SPICE files we may encounter
#define SUBCKTS  0x1	/* File contains .subckt definitions */
#define GENCALLS 0x2	/* File contains only X record calls */
#define FETONLY  0x4	/* File contains only M record MOSFETS */
#define OTHER	 0x8	/* File contains device types other than X or M */

/*--------------------------------------------------------------*/
/* Structures and Type Definitions				*/
/*--------------------------------------------------------------*/

/* Linked list of contraints */

typedef struct _constraint *constraintptr;
typedef struct _constraint {
   int type;		// LEFT, RIGHT, TOP, BOTTOM, or TOGETHER (see above)
   int thisnet;		// number of the constrained net;
   int thatnet;		// second net number for "together" constraint.
   int allow;		// Allow this much slack before applying cost
   int weight;		// Weight the cost function by this factor.
   constraintptr next;  // next in linked list
} constraint;

typedef struct _net {
   int metal;		// type of metal used (only used when doing layout)
   int routed;		// temp. flag for checking if net has been processed
   int channel;		// channel to which the net belongs
   int begin;		// min. x value in channel
   int end;		// max. x value in channel
   char *name;		// name of the network
} net, *netptr;

typedef struct _point {
   int x;
   int y;
} point;

typedef struct _vsegment {
   int net;		// net number of the connection
   int metal;		// type of metal used (only used when doing layout)
   point l;		// lower connection of the vertical segment
   point u;		// upper connection of the vertical segment
   point contact;	// position of contact to channel (if any)
   int yjog;		// vertical position of a jog (if any)
} vsegment, *vsegmentptr;

typedef struct _cell {
   int order;	// cell number as ordered in the source file; blanks are order=0
   int orient;	// for fet cell orient = 0 is GDS, orient = 1 is SDG (flipped)
   int overlap;	// cell position can be made tighter by overlapping next cell
   int *routed; // list of flags per port indicating which nets have been routed
   int *net;	// list of nets per port
   point *pos;  // x and y position of cell
   int boxnum;	// boxarr[boxnum] gives info about cell's type
} cell, *cellptr;

/* Enough information to reconstruct a cell from saved state */

typedef struct _cellinfo {
   cellptr cell;
   int orient;
} cellinfo, *cellinfoptr;

/* Enough information to reconstruct the layout from a saved state */

typedef struct _info {
   cellinfo *upper;
   cellinfo *lower;
   int cost;
   int channels;
   int rlength;
   int misalign;
} info, *infoptr;

/* Check for routability to each side of a port */
/* (One for horizontal, one vertical)		*/

/* Check for routability to each side of a port */
/* (One for horizontal, one vertical)           */

typedef struct _routepart {
   int min;
   int max;
} routepart, *routepartptr;

typedef struct _routestruct {
   routepartptr routes;
   int numroutes;
} routestruct;

typedef struct _portroute {
   routestruct ns;
   routestruct ew;
} portroute, *portrouteptr;

typedef struct _port {
   char *name;
   point ll;    /* lower left (route always starts from here) */  
   int north;   /* 1 if port can be routed to the north, 0 otherwise */
   int south;
   int east;
   int west;
   int idx;		// Stores order in which nets are read from spice file
   			// if spice[j - 1] = i, net connected to "ports[i]"
   			// is the jth one to be read in.
} portstruct, *portstructptr; 

// Caveats on portstruct "idx" record:  If the magic "port" record is used,
// then the index declared by the "port" record is used, and a port ordering
// is not required from SPICE (that is, the SPICE deck containing those
// cells does not need to declare subcircuits).  BUT, for MOSFET cells, the
// port ordering must be DRAIN, GATE, SOURCE, [BULK].  If port records
// exist in a MOSFET cell, they must match this order.

typedef struct _rect {
   point ll;	/* lower left */
   point ur;	/* upper right */
} rect, *rectptr;

typedef struct _box {
   char *name;
   int ht; /* height */
   int w;  /* width */
   int x;  /* x offset */
   int y;  /* y offset */
   int instance;   /* instance number */
   int count;  /* total number of cells of this type */
   int numports;   // number of ports in cell type
   portstructptr ports; 
   rect gnd;
   rect vdd;
} box, *boxptr;

typedef struct _portinfo {
   cellptr *source;
   int pos;
   int port;
} portinfo, *portinfoptr;

typedef struct _vsave {
   point endpt;
   int row;
   int net;
   int jogy;
   int jogx;
} vsave;

typedef struct _m1save {
   int net1;	/* net of last vertical line without jumper	*/
   int net2;	/* net of 2nd to last line without jumper	*/
   int x1;	/* position of last line			*/
   int x2;	/* position of 2nd to last line			*/
   int y1;	/* endpoint of last line			*/
   int y2;	/* endpoint of 2nd to last line			*/
   int row1;	/* row of last port				*/
   int row2;	/* row of 2nd to last port			*/
} m1save;

/* Each channel is a list of networks	 		*/
/* with an additional integer denoting the number	*/
/* of row2-connections from the channel less the number	*/
/* of row1-connections:  indicates where the net goes	*/
/* best vertically.					*/

typedef struct _channel {
   int lucount;			// (see above)
   int num_nets;		// number of networks occupying the channel
   int *nets;			// list of networks occupying the channel
   int yval;			// position of channel (after sorting)
} chstruct, *chstructptr;

/*--------------------------------------------------------------*/
/* Header files							*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <values.h>
#include <termios.h>
#include <sys/ioctl.h>

/*--------------------------------------------------------------*/
/* Global Variables						*/
/*--------------------------------------------------------------*/

extern int netcount;
extern netptr networks;

// lower and upper cells (to replace ncells and pcells, respectively)
extern cellptr *lcells, *ucells;
extern int cols;
extern char placeinfo[385];
extern boxptr boxarr;
extern int boxcount;
extern int fet_design;

// from gnd_vdd.c

extern int vdd_line;
extern int gnd_line;
extern int vddnum, gndnum, sprnum;
extern cellptr *sprcells;
extern int sprcount;
extern int first_gv;

/*--------------------------------------------------------------*/
/* Function Prototypes						*/
/*--------------------------------------------------------------*/

// from router-1.4.c

int chcompare(chstructptr a, chstructptr b);
void setnormal(int fd);

// from gnd_vdd.c

int gv_ptp(int, int, cellptr *, int, int);
cellptr find_closest_gv(int, cellptr, int);
netptr create_gvnet(int, int);
int vddcellcmp(cellptr *, cellptr *);
int gndcellcmp(cellptr *, cellptr *);
void convert_to_gv(cellptr, int, int);
cellptr find_intermediate_cell(int, int, int);
void gvroute(int);
void draw_gv_ptp(FILE *, cellptr, cellptr);
