/*----------------------------------------------------------------------*/
/* genrouter.c								*/
/*									*/
/* SPICE-file genetic algorithm-based place-and-route:	    		*/
/*									*/
/* This is a channel router for standard cells.  It can route 		*/
/* either transistor designs, assuming that all n-devices lie in	*/
/* a row and all p-devices lie in a row, or subcircuit designs that 	*/
/* route smaller cells.  The router uses a genetic algorithm to search	*/
/* for a placement solution that uses, as its fitness function, the	*/
/* total route length required to make all necessary connections.	*/
/* Routing is determined by the right-hand-rule (greedy) algorithm.	*/
/* The genetic algorithm uses the ordered placement and orientation of	*/
/* cells as its	"gene".  The mutation method exchanges two randomly	*/
/* selected cells or groups of cells, either on the top row,  bottom	*/
/* row, or both, and possibly flipped.  On each cycle, repeated genes	*/
/* are culled from the population, leaving unique genes only, which are	*/
/* sorted according to fitness score.  The population is redistributed	*/
/* with representation in the population computed as a function of	*/
/* fitness, such that more fit individuals have a greater		*/
/* representation.  Mutations are applied with a given mutation rate,	*/
/* and mutated genes are evaluated for fitness.  The evolutionary cycle	*/
/* is an infinite loop which is stopped by the user.  Upon stopping,	*/
/* the program generates the .mag layout file corresponding to the	*/
/* best-known layout.							*/
/*									*/
/* version 1.5:   Adapted to read .spc and .fspc (spice and		*/
/* flattened-spice) input files and to route subcircuit designs.	*/
/* Added capability to read a constraints file, if one exists.		*/
/*									*/
/* version 1.6:   Changed to a "genetic algorithm" version to		*/
/* overcome limitations of settling into local minimum solutions.	*/
/*									*/
/* version 1.7:	  Changed to graphic output on an X window and		*/
/* in a PostScript output file, replacing the ASCII text version.	*/
/*									*/
/* version 1.8:	  Reworked the vertical routing into a "wire"-type	*/
/* structure.								*/
/*									*/
/* version 1.9:  Revised code based on the new "port"-type labels in	*/
/* magic version 7.2.							*/
/*									*/
/* Copyright 2001, 2003 - Jennifer Roberts and Tim Edwards		*/	
/*----------------------------------------------------------------------*/

#include "defines.h"

#include <math.h>

/*----------------------*/
/* Global variables	*/
/*----------------------*/

int netcount;
netptr networks;
constraintptr constraints;
info best;
char wait;

int ch_minimum;		   // target minimum number of channels
int ch_penalty;		   // cost per channel per cell in excess of target
int align_penalty;	   // cost of misalignment between rows
int m1_width;		   // width of a m1 line
int m2_width;		   // width of a m2 line
int via_width;		   // width of a via
int via_height;		   // height of a via
int m1_space;		   // spacing between m1 lines
int m2_space;		   // spacing between m2 lines
int contact_size;	   // contact size for a substrate/well contact

cellptr *lcells, *ucells;  // lower and upper cells (to replace ncells and 
      			   // pcells, respectively)
int cols;
char placeinfo[385];
char techname[30];
boxptr boxarr = NULL;
int boxcount = 0;
int fet_design;
portinfoptr portorder;
int total_ports;


extern int open_window();
extern point *matching_v(int, int, cellptr *, int);
extern boxptr read_box(char *);
extern void scalecoord(point *, int, int);

/*----------------------------------------------*/
/* Close graphics window and quit		*/
/*----------------------------------------------*/

void quit(int status)
{
   close_window();
   exit(status);
}

/*----------------------------------------------*/
/* qsort() routines:				*/
/*----------------------------------------------*/

/*----------------------------------------------*/
/* Compare function for sorting solutions	*/
/* (reverse sort:  highest to lowest score)	*/
/*----------------------------------------------*/

int infocompare(infoptr a, infoptr b)
{
   if (b->cost == a->cost) return 0;
   else return (b->cost > a->cost) ? -1 : 1;
}

/*-----------------------------------------------------------*/
/* Compare function for sorting horizontal routes vertically */
/*-----------------------------------------------------------*/

int chcompare(chstructptr a, chstructptr b)
{
   if (b->lucount == a->lucount) return 0;
   else return (b->lucount > a->lucount) ? 1 : -1;
}

/*------------------------------------------------------*/
/* Compare function for ordering ports left-to-right	*/
/*------------------------------------------------------*/

int portcompare(portstructptr a, portstructptr b)
{
    if (b->ll.x == a->ll.x) return 0;
    else return (b->ll.x > a->ll.x) ? -1 : 1;
}

/*----------------------------------------------*/
/* Compare two portinfos to rank left-to-right	*/
/*----------------------------------------------*/

int picompare(portinfoptr a, portinfoptr b)
{
    cellptr acell = *(a->source + a->pos);
    cellptr bcell = *(b->source + b->pos);
    int ax = acell->pos[a->port].x;
    int bx = bcell->pos[b->port].x;

    if (bx == ax) return 0;
    else return (bx > ax) ? -1 : 1;
}

/*----------------------------------------------*/
/* Return terminal to normal settings           */
/*----------------------------------------------*/

void setnormal(int fd)
{
   struct termios tios;
 
   tcgetattr(fd, &tios);
   tios.c_lflag |= ICANON | ECHO;
   tcsetattr(fd, TCSANOW, &tios);
}

/*----------------------------------------------*/
/* Set input to return characters immediately   */
/* and without echoing to the screen            */
/*----------------------------------------------*/

void setnodelay(int fd)
{  
   struct termios tios;
   
   tcgetattr(fd, &tios); 
   tios.c_lflag &= (~(ICANON | ECHO));
   tios.c_cc[VMIN] = 0;
   tios.c_cc[VTIME] = 0; 
   tcsetattr(fd, TCSANOW, &tios);
}  

/*--------------------------------------------------------------*/
/* Seed the random number generator				*/
/*--------------------------------------------------------------*/

int setrandom(int start_time) {

   if (start_time != 0) {
      srandom(start_time);
      return start_time;
   }
   else {
      struct timeval ltv;
      struct timezone ltz;

      gettimeofday(&ltv, &ltz);
      srandom((int)ltv.tv_usec);
      return (int)ltv.tv_usec;
   }
}

/*--------------------------------------------------------------*/
/* Copy architecture information from one record to another	*/
/*--------------------------------------------------------------*/

void icopyfwd(infoptr from, infoptr to)
{
   int i;
   cellinfoptr ci, co;

   for (i = 0; i < cols; i++) {
      ci = from->upper + i;
      co = to->upper + i;
      co->cell = ci->cell;
      co->orient = ci->orient;

      ci = from->lower + i;
      co = to->lower + i;
      co->cell = ci->cell;
      ci->orient = co->orient;
   }
   to->cost = from->cost;
   to->channels = from->channels;
   to->rlength = from->rlength;
   to->misalign = from->misalign;
}

/*--------------------------------------------------------------*/
/* Save architecture information				*/
/*--------------------------------------------------------------*/

void ccopyfwd(cellptr *upper, cellptr *lower, infoptr archinfo)
{
   int i;
   cellinfoptr ci;

   for (i = 0; i < cols; i++) {
      ci = archinfo->upper + i;
      ci->cell = *(upper + i);
      ci->orient = (*(upper + i))->orient;

      ci = archinfo->lower + i;
      ci->cell = *(lower + i);
      ci->orient = (*(lower + i))->orient;
   }
}

/*--------------------------------------------------------------*/
/* Recall architecture information				*/
/*--------------------------------------------------------------*/

void ccopyback(infoptr archinfo, cellptr *upper, cellptr *lower)
{
   int i;
   cellinfoptr ci;

   for (i = 0; i < cols; i++) {
      ci = archinfo->upper + i;
      *(upper + i) = ci->cell;
      (*(upper + i))->orient = ci->orient;

      ci = archinfo->lower + i;
      *(lower + i) = ci->cell;
      (*(lower + i))->orient = ci->orient;
   }
}

/*--------------------------------------------------------------*/
/* Remove the newline character from the end of a string	*/
/*--------------------------------------------------------------*/

remove_nl(char *line_in)
{
   char *sptr = line_in;
   while (*(++sptr) != '\n')
      if (*sptr == '\0') return; 
   *sptr = '\0';
}

/*--------------------------------------------------------------*/
/* check_net:  Find a network by name.  If NULL is passed as	*/
/* net_str, then an internal name is generated and passed to	*/
/* the network.							*/
/*--------------------------------------------------------------*/

int check_net(char const *net_str)
{
   netptr newnet;
   int i;
   static int inode = 0;

   if (net_str != NULL) {
      for (i = 0; i < netcount; i++) {
	 if (!strcmp(net_str, (networks + i)->name)) return (i);
      }
   }

   netcount++;
   networks = (netptr)realloc(networks, netcount * sizeof(net));
   newnet = networks + netcount - 1;
   newnet->name = (char *)malloc((1 + strlen(net_str)) * sizeof(char)); 

   if (net_str != NULL)
      strcpy(newnet->name, net_str);
   else
      sprintf(newnet->name, "inode%d", inode++);

   return (netcount - 1);
}

/*--------------------------------------------------------------*/
/* find_matching_box ---					*/
/*								*/
/*	Find the box which matches the name.  If no box is	*/
/*	found, then try to load a SPICE deck corresponding	*/
/* 	to the name.  If all else fails, return -1.		*/
/*--------------------------------------------------------------*/

int find_matching_box(char *name)
{
   int i;
   boxptr newbox;
   
   for (i = 0; i < boxcount; i++)
      if (strcmp(name, boxarr[i].name) == 0)
	 return i;  // return corresponding boxnum

   newbox = read_box(name);
   return (newbox == NULL) ? -1 : (boxcount - 1);
}

/*----------------------------------------------------------------------*/
/* allocate memory for a new cell in the ucell or lcell array;		*/
/* allocate memory for the new cell's routed, net, and pos arrays.	*/
/* initialize cell datamembers and update corresponding boxarr element	*/
/* rownum = 0 corresponds to bottom row, cellnum = cell# on that row	*/
/*----------------------------------------------------------------------*/

void create_cell(int boxnum, int rownum, int cellnum, char **nets_in)
{
   cellptr *newcellptr, newc;
   int i, portcount = boxarr[boxnum].numports;
   boxptr bx = boxarr + boxnum;

   if (rownum)
      ucells = (cellptr *)realloc(ucells, cellnum * sizeof(cellptr));
   else
      lcells = (cellptr *)realloc(lcells, cellnum * sizeof(cellptr));

   newcellptr = ((rownum ? ucells : lcells) + cellnum - 1);
   *(newcellptr) = (cellptr)malloc(sizeof(cell));
   
   newc = *newcellptr;
   if (nets_in[0] != '\0'  && portcount != 0) {
      newc->routed = (int *)malloc(portcount * sizeof(int));
      newc->net = (int *)malloc(portcount * sizeof(int));
      newc->pos = (point *)malloc(portcount * sizeof(point));
      // store input nets into net array
      for (i = 0; i < portcount; i++) {
         newc->net[i] = check_net(nets_in[(boxarr + boxnum)->ports[i].idx]);
         newc->routed[i] = 0;
      }
   }
   else {
      newc->routed = newc->net = NULL;
      if (bx->w == 0 || (!vdd_line && !gnd_line))
         newc->pos = NULL;
      else {
	 newc->pos = (point *)malloc(2 * sizeof(point));
	 newc->net = (int *)malloc(2 * sizeof(int));
	 newc->routed = (int *)malloc(2 * sizeof(int));
	 newc->net[GND] = GND;
	 newc->net[VDD] = VDD;
	 newc->routed[GND] = newc->routed[VDD] = 0;
	 // in subcircuit designs that have gnd/vdd connections, log position
	 // of gnd/vdd/spacer cells.
	 if ((++sprcount) == 1)
	    sprcells = (cellptr *)malloc(sizeof(cellptr));
	 else
	    sprcells = (cellptr *)realloc(sprcells, sizeof(cellptr) * sprcount);
	 *(sprcells + sprcount - 1) = *newcellptr;
      }
   }
   
   newc->order = newc->orient = newc->overlap = 0;
   newc->boxnum = boxnum;
   bx->count++;
}

/*--------------------------------------------------------------*/
/* Sort port directional routes					*/
/*--------------------------------------------------------------*/

int rpcompare(routepartptr a, routepartptr b)
{
   if (a->min > b->min)
      return 1;
   else if (a->min < b->min)
      return -1;
   else {
      if (a->max > b->max) return 1;
      else if (a->max < b->max) return -1; 
      else return 0;
   }
}

/*--------------------------------------------------------------*/
/* For each port, search through magic file to check whether	*/
/* the port can be routed north, south, east, or west		*/
/* Assume all fet_design routing is with metal1 and all other	*/
/* subcell designs use metal2 routing until outside of the cell.*/
/* Assume the only metal2 going east-west in non_fet_designs	*/
/* are jogs (rectangles saved in one piece).  Assume only	*/
/* vertical metal2 lines are stored in multiple pieces		*/
/*--------------------------------------------------------------*/

void check_nsew(FILE *file, boxptr nb, int needscale, int scalen, int scaled)
{
   char line_in[100], *metal = (fet_design ? "metal1" : "metal2");
   int i, j, k, numparts;
   rect temp;
   int lower1, upper1, lower2, upper2;
   portrouteptr proutes = NULL;
   routepartptr nsroutes, ewroutes, endpart;
   
   if (nb->numports > 0) {
      proutes = (portrouteptr)malloc(nb->numports * sizeof(portroute));
   }
   for (i = 0; i < nb->numports; i++) {
      proutes[i].ns.numroutes = 0;
      proutes[i].ns.routes = (routepartptr) malloc(sizeof(routepart));
      proutes[i].ew.numroutes = 0;
      proutes[i].ew.routes = (routepartptr) malloc(sizeof(routepart));
   }

   rewind(file);
   while (fgets(line_in, 99, file) != NULL) {
      if (strstr(line_in, metal) != NULL || strstr(line_in, "via") != NULL
	    || strstr(line_in, "m2contact") != NULL ||
	    (fet_design && strstr(line_in, "contact") != NULL)) {
	 while (fgetc(file) == 'r' && fgets(line_in, 99, file) != NULL &&
	       strstr(line_in, "label") == NULL) {
	    sscanf(line_in, "%*s %d %d %d %d", &temp.ll.x, &temp.ll.y,
		  &temp.ur.x, &temp.ur.y);
	    // scale label position coordinate if necessary
	    if (needscale) {
	       scalecoord(&temp.ll, scalen, scaled);
	       scalecoord(&temp.ur, scalen, scaled);
	    }
	    for (i = 0; i < nb->numports; i++) {
	       /* case E-W */
	       if (temp.ll.y < nb->ports[i].ll.y + m1_width &&
		   	temp.ur.y > nb->ports[i].ll.y) {
		  numparts = ++proutes[i].ew.numroutes;
		  proutes[i].ew.routes = (routepartptr) realloc(
				proutes[i].ew.routes,
				numparts * sizeof(routepart));
		  proutes[i].ew.routes[numparts - 1].min = temp.ll.x;
		  proutes[i].ew.routes[numparts - 1].max = temp.ur.x;
	       }

	       /* case N-S */
	       if (temp.ll.x < nb->ports[i].ll.x + m1_width &&
		   	temp.ur.x > nb->ports[i].ll.x) {
		  numparts = ++proutes[i].ns.numroutes;
		  proutes[i].ns.routes = (routepartptr) realloc(	
				proutes[i].ns.routes,
				numparts * sizeof(routepart));
		  proutes[i].ns.routes[numparts - 1].min = temp.ll.y;
		  proutes[i].ns.routes[numparts - 1].max = temp.ur.y;
	       }
	    }
	 }
      }
   }
   
   // sort rect's in portstruct arrays from smallest ll.x/y to largest ll.x/y
   for (i = 0; i < nb->numports; i++) {
      nsroutes = proutes[i].ns.routes;
      ewroutes = proutes[i].ew.routes;
      qsort(nsroutes, proutes[i].ns.numroutes, sizeof(routepart), rpcompare);
      qsort(ewroutes, proutes[i].ew.numroutes, sizeof(routepart), rpcompare);

      // merge overlapping rects N-S
     
      j = 0;
      while (j < proutes[i].ns.numroutes - 1) {
	 if (nsroutes[j + 1].min <= nsroutes[j].max &&
	     	nsroutes[j + 1].max >= nsroutes[j].min) {
	    nsroutes[j].min = MIN(nsroutes[j].min, nsroutes[j + 1].min);
	    nsroutes[j].max = MAX(nsroutes[j].max, nsroutes[j + 1].max);
	    // shift the rest of the route parts down by one
	    proutes[i].ns.numroutes--;
	    for (k = j + 1; k < proutes[i].ns.numroutes; k++) {
	       nsroutes[k].max = nsroutes[k + 1].max;
	       nsroutes[k].min = nsroutes[k + 1].min;
	    }
	 }
	 else j++;
      }

      // merge overlapping rects E-W

      j = 0;
      while (j < proutes[i].ew.numroutes - 1) {
	 if (ewroutes[j + 1].min <= ewroutes[j].max &&
	     	ewroutes[j + 1].max >= ewroutes[j].min) {
	    ewroutes[j].min = MIN(ewroutes[j].min, ewroutes[j + 1].min);
	    ewroutes[j].max = MAX(ewroutes[j].max, ewroutes[j + 1].max);
	    // shift the rest of the route parts down by one
	    proutes[i].ew.numroutes--;
	    for (k = j + 1; k < proutes[i].ew.numroutes; k++) {
	       ewroutes[k].max = ewroutes[k + 1].max;
	       ewroutes[k].min = ewroutes[k + 1].min;
	    }
	 }
	 else j++;
      }

      // determine routability S

      nb->ports[i].south = (nsroutes->max >= nb->ports[i].ll.y) ? 1 : 0;

      // determine routability N

      endpart = nsroutes + proutes[i].ns.numroutes - 1;
      nb->ports[i].north = (endpart->min <= nb->ports[i].ll.y) ? 1 : 0;

      // determine routability W

      nb->ports[i].west = (ewroutes->max >= nb->ports[i].ll.x) ? 1 : 0;

      // determine routability E

      endpart = ewroutes + proutes[i].ew.numroutes - 1;
      nb->ports[i].east = (endpart->min <= nb->ports[i].ll.x) ? 1 : 0;

      // clean up

      free(proutes[i].ns.routes);
      free(proutes[i].ew.routes);
   }
   if (nb->numports > 0) free(proutes);

   // check to see whether e/w routes will interfere with n/s routes

   for (i = 0; i < (nb->numports - 1); i++) {
      for (j = (i + 1); j < nb->numports; j++) {
	 if (nb->ports[i].ll.x < nb->ports[j].ll.x) {
	    if (nb->ports[i].ll.y < nb->ports[j].ll.y) {
	       nb->ports[i].east = (nb->ports[i].east ? IF_ON_BOTTOM : 0);
	       nb->ports[j].west = (nb->ports[j].west ? IF_ON_TOP : 0);
	    }
	    else {
	       nb->ports[i].east = (nb->ports[i].east ? IF_ON_TOP : 0);
	       nb->ports[j].west = (nb->ports[j].west ? IF_ON_BOTTOM : 0);
	    }
	 }
	 else if (nb->ports[i].ll.x > nb->ports[j].ll.x) {
	    if (nb->ports[i].ll.y < nb->ports[j].ll.y) {
	       nb->ports[i].west = (nb->ports[i].west ? IF_ON_BOTTOM : 0);
	       nb->ports[j].east = (nb->ports[j].east ? IF_ON_TOP : 0);
	    }
	    else {
	       nb->ports[i].west = (nb->ports[i].west ? IF_ON_TOP : 0);
	       nb->ports[j].east = (nb->ports[j].east ? IF_ON_BOTTOM : 0);
	    }
	 }
      }
   }

   /* Diagnostic */
   // fprintf(stderr, "Routability of pins in cell %s:\n", nb->name);
   // for (i = 0; i < nb->numports; i++)
   //    fprintf(stderr, "Pin %d (\"%s\"): N=%d S=%d E=%d W=%d\n",
   //		i, nb->ports[i].name, nb->ports[i].north,
   //		nb->ports[i].south, nb->ports[i].east,   nb->ports[i].west);
}

/*--------------------------------------------------------------*/
/* Coordinate scaling when "magscale" line is present		*/
/*--------------------------------------------------------------*/

void scalecoord(point *p, int scalen, int scaled)
{
   p->x = (p->x * scalen) / scaled;
   p->y = (p->y * scalen) / scaled;
}

/*--------------------------------------------------------------*/
/* Read magic file corresponding to cell in spice file		*/
/*--------------------------------------------------------------*/

boxptr read_box(char *name)
{
   FILE *file;
   char *filename, portname;
   char linein[100], *lineptr, stemp[50], label[50], contact[49];
   point max, min;
   rect bx;
   int i, j, k, transform[3][2], scalen, scaled, needscale;
   portstructptr newport;
   boxptr newbox;
   int minidx, pindex, drawn_ports = 0;

   /* The order of fetnames[] matches the order of an "M" record in a	*/
   /* SPICE file.							*/

   static char *fetnames[] = {"dD", "gG", "sS", "bBwW"};

   filename = (char *)malloc((strlen(name) + 5) * sizeof(char));
   strcpy(filename, name);
   strcat(filename, ".mag");

   if ((file = fopen(filename, "r")) == NULL) {
      fprintf(stderr, "Can't open file \"%s\" for reading\n", filename);
      if (strstr(name, "cap") != NULL || (!vdd_line && !gnd_line &&
	       strstr(name, "spacer") != NULL))
	 return NULL;
      else
         quit(1);
   }
   fprintf(stderr, "Reading layout of cell %s.\n", name);

   /* allocate memory for new box */
   boxcount++;
   boxarr = (boxptr)(boxcount == 1 ? malloc(sizeof(box)) : realloc(boxarr,
	    sizeof(box) * boxcount));
   newbox = boxarr + (boxcount - 1);
   
   /* store name of box */
   newbox->name = (char *)malloc((1 + strlen(name)) * sizeof(char));
   strcpy(newbox->name, name);
   if (strcmp(name, "Vdd") == 0 || strcmp(name, "GND") == 0)
      newbox->name = (char *)realloc(newbox->name, (1 + strlen("spacer"))
	    * sizeof(char));

   /* read in box info (ports, GND, VDD, box height and width) */

   newbox->numports = 0;
   max.x = max.y = 0;
   min.x = MAX_BOX_WIDTH;
   min.y = MAX_BOX_HEIGHT;

   needscale = 0;

   /* skip header to end of "timestamp" line, making sure that the	*/
   /* technology matches what is expected.				*/

   while (1) {
      if (fgets(linein, 100, file) == NULL) {
	 fprintf(stderr, "Error:  Layout file has no timestamp.  Aborting!\n");
	 quit(1);
      }
      remove_nl(linein);

      /* Make sure technology is correct */

      if (!strncmp(linein, "tech", 4)) {
	 if (techname[0] == '\0')
	    strncpy(techname, linein + 5, 29);
	 else
	    if (strcmp(techname, linein + 5))
	       fprintf(stderr, "***Warning!***  Tech name \"%s\" differs from "
			"orignal name \"%s\"!\n", linein + 5, techname);
      }

      /* Magscale must be handled correctly */

      else if (!strncmp(linein, "magscale", 8)) {
	 if (sscanf(linein, "%*s %d %d", &scalen, &scaled) != 2)
	 {
	    fprintf(stderr, "Warning:  badly formed magscale line in header!\n");
	    needscale = 0;
	 }
	 else needscale = 1;
      }

      /* Skip over timestamp, scale, etc. */

      else if (!strncmp(linein, "timestamp", 9))
	  break;
   }


   while (fgets(linein, 100, file) != NULL) {
      remove_nl(linein);

      if (strcmp("<< labels >>", linein) == 0) {
	 newport = NULL;
	 while(fgets(linein, 100, file) != NULL) {
	    remove_nl(linein);
	    if (!strncmp(linein, "<<", 2))
	       break;
	    else if (!strncmp(linein, "port", 4)) {
	       // port record is supposed to relate to label on previous
	       // line, so "newport" should still point to the port record
	       if (newport == NULL) {
		  fprintf(stderr, "Bad port record:  Port does not refer to label!\n");
		  fprintf(stderr, "Offending line is: \"%s\"\n", linein);
		  continue;
	       }
	       else {
		  if (sscanf(linein, "%*s %d", &pindex) != 1) {
		     fprintf(stderr, "Bad port record in layout.\n");
		     continue;
		  }
		  drawn_ports = 1;
		  newport->idx = pindex;
		  newport = NULL;
	       }
	    }
	    else if (sscanf(linein, "%*s %49s %d %d %d %d %*d %49s", contact, 
		&bx.ll.x, &bx.ll.y, &bx.ur.x, &bx.ur.y, label) < 6) {
	       fprintf(stderr, "Error in reading label from %s:", filename);
	       fprintf(stderr, "Offending line is: \"%s\"\n", linein);
	    }
	    else {  // regular label (not a port record or a GND/Vdd connection)
	       (newbox->numports)++;
	       newbox->ports = (portstructptr)(newbox->numports == 1 ? 
		     malloc(sizeof(portstruct)) : realloc(newbox->ports,
		     newbox->numports * sizeof(portstruct)));
	       newport = newbox->ports + newbox->numports - 1;

	       // scale label position coordinate if necessary
	       if (needscale) {
		  scalecoord(&bx.ll, scalen, scaled);
		  scalecoord(&bx.ur, scalen, scaled);
	       }

	       // set the index of the port
	       newport->idx = -1;

	       // initialize ll's (lower left positions) in ports array
	       newport->ll = bx.ll;

	       // initialize names in ports array
	       newport->name = (char *)malloc((1 + strlen(label)) * sizeof(char));
	       strcpy(newport->name, label);

	       // initialize routing info to allow routing in all directions
	       newport->north = newport->south = newport->east = newport->west = 1;

	       // if this is a Vdd or GND connection, record it
	       if (!strcmp(label, "GND"))
	          newbox->gnd = bx;
	       else if (!strcmp(label, "VDD") || !strcmp(label, "Vdd"))
	          newbox->vdd = bx;
	    }
	 }
      }
      else if (linein[0] == '<')
	 continue;
      else {  // find lower left and upper right corner of box
	 sscanf(linein, "%49s %d %d %d %d", stemp, &bx.ll.x, &bx.ll.y, &bx.ur.x,
	       &bx.ur.y);
	 if (strcmp(stemp, "use") == 0) { // subcell present in .mag file
	    fgets(linein, 100, file);  // skip box timestamp line
	    fgets(linein, 100, file);  // box transform line
	    sscanf(linein, "%*s %d %d %d %d %d %d", &transform[0][0],
	          &transform[1][0], &transform[2][0], &transform[0][1],
		  &transform[1][1], &transform[2][1]);
	    fgets(linein, 100, file);
	    sscanf(linein, "%*s %d %d %d %d", &bx.ll.x, &bx.ll.y,
	    	   &bx.ur.x, &bx.ur.y);
	    bx.ll.x = transform[0][0] * bx.ll.x + transform[1][0] * bx.ll.y
	       + transform[2][0];
	    bx.ll.y = transform[0][1] * bx.ll.x + transform[1][1] * bx.ll.y
	       + transform[2][1];
	    bx.ur.x = transform[0][0] * bx.ur.x + transform[1][0] * bx.ur.y
	       + transform[2][0];
	    bx.ur.y = transform[0][1] * bx.ur.x + transform[1][1] * bx.ur.y
	       + transform[2][1];
	    if (bx.ll.x > bx.ur.x) {	// happens when box is flipped
	       i = bx.ll.x;
	       bx.ll.x = bx.ur.x;
	       bx.ur.x = i;
	    }
	    if (bx.ll.y > bx.ur.y) {
	       i = bx.ll.y;
	       bx.ll.y = bx.ur.y;
	       bx.ur.y = i;
	    }
	 }

	 /* scale coordinates */
	 if (needscale) {
	    scalecoord(&bx.ll, scalen, scaled);
	    scalecoord(&bx.ur, scalen, scaled);
	 }

	 /* determine height and width of cell */
	 if (bx.ll.x < min.x)  min.x = bx.ll.x;
	 if (bx.ll.y < min.y)  min.y = bx.ll.y;
	 if (bx.ur.x > max.x)  max.x = bx.ur.x;
	 if (bx.ur.y > max.y)  max.y = bx.ur.y;
      }
   }

   // initialize width and height
   newbox->w  = max.x - min.x;
   newbox->ht = max.y - min.y;

   // routing algorithm assumes that the box is positioned at zero;
   // we subtract the x and y offsets from the port positions to
   // compensate, and will add the offsets back in when positioning
   // the cells in the final output layout.

   newbox->x  = min.x;
   newbox->y  = min.y;

   // If we have encountered port statements (drawn_ports = 1), then all
   // ports with index -1 should be removed.

   if (drawn_ports) {
      for (i = 0; i < newbox->numports; i++) {
         if (newbox->ports[i].idx == -1) {
	    j = newbox->numports - 1;
	    free(newbox->ports[i].name);
	    newbox->ports[i].name = newbox->ports[j].name;
	    newbox->ports[i].ll = newbox->ports[j].ll;
	    newbox->ports[i].north = newbox->ports[j].north;
	    newbox->ports[i].south = newbox->ports[j].south;
	    newbox->ports[i].east = newbox->ports[j].east;
	    newbox->ports[i].west = newbox->ports[j].west;
	    newbox->ports[i].idx = newbox->ports[j].idx;
	    newbox->numports--;
	    i--;
	 }
      }

      // Ports indices should be ordered contiguously beginning from zero.
      // Magic does not require this, so we enforce it here.

      for (j = 0; j < newbox->numports; j++) {
	 minidx = 100000;
	 for (i = 0; i < newbox->numports; i++) {
	    if ((newbox->ports[i].idx >= j)
			&& (newbox->ports[i].idx < minidx)) {
	       minidx = newbox->ports[i].idx; 
	       pindex = i;
	    }
	 }
	 newbox->ports[pindex].idx = j;
      }
   }


   /* Sort ports from left to right so that leftmost port appears first	*/
   /* in array and rightmost appears last.				*/

   qsort(newbox->ports, newbox->numports, sizeof(portstruct), portcompare);

   /* port order is predetermined by SPICE format if routing an nfet/pfet design */

   if (fet_design && strstr(name, "cap") == NULL && strstr(name, "spacer")
	 == NULL) {
      if (newbox->numports != FETPORTS - 1 && newbox->numports != FETPORTS) {
	 fprintf(stderr, "Warning: (%cfet.mag).  Cell has %d ports.\nCell "
	       "should have %d or %d ports (not counting GND and VDD)."
	       "\n", name[0], newbox->numports, FETPORTS - 1, FETPORTS);
      }

      if (!drawn_ports) {
         for (i = 0; i < newbox->numports; i++) {
	    newbox->ports[i].idx = -1;
	    for (j = 0; j < newbox->numports; j++) { // j = matching port in array
	       for (k = 0; (portname = fetnames[i][k]) != '\0'; k++) {
	          if (newbox->ports[j].name[0] == portname) {
	             newbox->ports[j].idx = i;
	             break;
	          }
	       }
	       if (portname != '\0') break;
	    }
	    if (j == newbox->numports) {
	       fprintf(stderr, "Ports in <%s> labeled incorrectly.  FET terminal "
		        "\"%s\" doesn't begin with any letter from the"
			" list: gGsSdDdBwW\n", filename, newbox->ports[j].name);
	       quit(1);
	    }
	 }
      }
   }

   // initialize instance and count data members
   newbox->instance = newbox->count = 0;
   
   // check to see which direction ports can be routed in
   check_nsew(file, newbox, needscale, scalen, scaled);
   
   fclose(file);
   
   // Recenter all ports and Vdd/GND connection locations for a
   // cell origin of (0,0)

   for (i = 0; i < newbox->numports; i++) {
      newbox->ports[i].ll.x -= min.x;
      newbox->ports[i].ll.y -= min.y;
   }
   newbox->gnd.ll.x -= min.x;
   newbox->gnd.ll.y -= min.y;
   newbox->gnd.ur.x -= min.x;
   newbox->gnd.ur.y -= min.y;

   newbox->vdd.ll.x -= min.x;
   newbox->vdd.ll.y -= min.y;
   newbox->vdd.ur.x -= min.x;
   newbox->vdd.ur.y -= min.y;

   // initialize ports in vdd/gnd cells
   if (strcmp(name, "Vdd") == 0 || strcmp(name, "GND") == 0 ||
	 (strcmp(name, "spacer") == 0 && (vdd_line || gnd_line))) {
      newbox->numports = 2;
      newbox->ports = (portstructptr)malloc(2 * sizeof(portstruct));
      newbox->ports[GND].name = malloc(sizeof(char) * 4);
      newbox->ports[VDD].name = malloc(sizeof(char) * 4);
      strcpy(newbox->ports[GND].name, "GND");
      strcpy(newbox->ports[VDD].name, "VDD");
      for (i = 0; i < newbox->numports; i++) {
	 portstructptr port = newbox->ports + i;
	 port->east = port->west = port->north = port->south = 1;
      }
   }
   free(filename);

   return (newbox);
}

/*----------------------------------------------------------------------*/
/* Read in a hierarchial .spc file (with .subckt definitions)		*/
/*----------------------------------------------------------------------*/

int read_spice_subckts(FILE *fi, char line_in[])
{
   char sub_name[50], **ports_in = NULL, *sptr;
   int i, j, numports, maxports;
   boxptr bx;
   int class = 0;

   // read subcircuit info from spc file and read box info from matching
   // .mag file.  In this case, the SPICE deck determines (overrides) port
   // order.

   maxports = 0;
   do {
      if (strstr(line_in, ".subckt") != NULL) {
	 sscanf(line_in, "%*s %49s", sub_name);
	 class |= SUBCKTS;
	 bx = read_box(sub_name);
	 
	 numports = bx->numports;
	 if (maxports < numports)
	    ports_in = (char **)realloc(ports_in, sizeof(char *) * numports);

	 sptr = strstr(line_in, sub_name) + strlen(sub_name) + 1;

	 // read subcircuit ports from spice file

	 for (i = 0; i < numports; i++) {
	    bx->ports[i].idx = -1;
	    if (i > (maxports - 1))
	       *(ports_in + i) = (char *)malloc(sizeof(char) * 50);
	    if (sscanf(sptr, "%49s", *(ports_in + i)) == 0) {
	       fprintf(stderr, "Warning: %d ports in SPICE file, %d ports"
			" in layout for cell %s\n", i + 1, numports,
			sub_name);
	       break;
	    }

	    // move to next token (port) in SPICE subcircuit definition
	    sptr += strlen(*(ports_in + i)) + 1;

	    // set port order by comparing SPICE port names to layout port names
	    for (j = 0; j < numports; j++)
	       if (strcmp(*(ports_in + i), bx->ports[j].name) == 0)
		  bx->ports[j].idx = i;

	    if (j == numports) {
	       fprintf(stderr, "Warning: port %s in SPICE deck does not "
			"match any port names in the layout for cell %s\n",
			*(ports_in + i), sub_name); 
	    }

	 }
	 if (numports > maxports)  maxports = numports;
	 while (fgets(line_in, 99, fi) != NULL && strstr(line_in, ".ends")
	       == NULL);  // skip to end of subcell definition
      }
      else if (toupper(line_in[0]) == 'X')
	 class |= GENCALLS;
      else if (toupper(line_in[0]) == 'M')
	 class |= FETONLY;
   } while (fgets(line_in, 99, fi) != NULL);

   // deallocate memory

   for (i = 0; i < maxports; i++)
      free(*(ports_in + i));
   free(ports_in);

   // return status must be one of the types of files defined in defines.h

   switch (class) {
      case FETONLY:
      case GENCALLS:
	 return class;
	 break;
      case GENCALLS | SUBCKTS:
	 return SUBCKTS;
	 break;
      default:
	 return OTHER;
   }
}

/*----------------------------------------------------------------------*/
/* Read in a SPICE deck with general subcells (X records)		*/
/* Currently, this deals only with X records, although it should be	*/
/* expanded to deal with any mixture of devices.			*/
/*----------------------------------------------------------------------*/

int read_spice_general(FILE *fi, char line_in[])
{
   char *sptr;
   char **tokens;
   int i, maxtokens, numtokens;
   int row = 0, ucols = 0, lcols = 0, maxcols, boxnum;
   int ulength = 0, llength = 0, vddcount = 0, gndcount = 0;
   int j, k;
   cellptr *source;
   boxptr sp1, sp2, vddptr, gndptr, sprptr;

   maxtokens = 10;
   tokens = (char **)malloc(10 * sizeof(char *));
   for (i = 0; i < maxtokens; i++) tokens[i] = (char *)NULL;

   while (fgets(line_in, 99, fi) != NULL) {
      if (line_in[0] != 'X') continue;

      if (row == 0)
	 lcols++;
      else
	 ucols++;

      sptr = strchr(line_in, ' ') + 1;

      for (i = 0;; i++) {
	 if (i == maxtokens) {
	    maxtokens++;
	    tokens = (char **)realloc(tokens, maxtokens * sizeof(char *));
	 }
         tokens[i] = (char *)malloc(50 * sizeof(char));

	 if (sscanf(sptr, "%49s", *(tokens + i))) {
	    sptr += strlen(*(tokens + i)) + 1; // + 1 to skip space
	    if (strcmp(*(tokens + i), "Vdd") == 0)
	       vddcount++;
	    if (strcmp(*(tokens + i), "GND") == 0)
	       gndcount++;
	 }
	 if (*sptr == '\0' || *sptr == '\n') {
	    numtokens = i;
	    boxnum = find_matching_box(*(tokens + i)); // last string = name

	    if (boxarr[boxnum].numports != numtokens) {
	       fprintf(stderr, "Error: The number of ports in SPICE is %d, but "
			"in layout is %d for cell %s\n",
			numtokens, boxarr[boxnum].numports, *(tokens + i));
	       quit(1);
	    }

	    // /* Diagnostic */
	    //
	    // fprintf(stderr, "SPICE: Call to cell %s:  Ports are:\n",
	    // 		*(tokens + i));
	    // for (j = 0; j < numtokens; j++) {
	    //   fprintf(stderr, "Port %d: Net name is %s (SPC),",
	    //		j, *(tokens + j));
	    //
	    //   for (k = 0; k < numtokens; k++)
	    //	    if (boxarr[boxnum].ports[k].idx == j)
	    //	       break;
	    //   if (k == numtokens) {
	    //	    fprintf(stderr, " No such port in layout!\n");
	    //	    for (k = 0; k < numtokens; k++)
	    //	       fprintf(stderr, "   Port %d (MAG) index is %d\n",
	    //			k, boxarr[boxnum].ports[k].idx);
	    //   }
	    //   else
	    //	    fprintf(stderr, " Port name is %s (MAG)\n",
	    //		boxarr[boxnum].ports[k].name);
	    // }

	    create_cell(boxnum, row, lcols, tokens);
	    (*((row == 0 ? lcells : ucells) + lcols - 1))->order =
	       lcols + ucols;
	    if (row == 0)
	       llength += boxarr[boxnum].w;
	    else
	       ulength += boxarr[boxnum].w;
	    break;
	 }
      }
      row = (row == 0);	// flip flop row to put 1st cell on row 0, 2nd on row 1
   }
   
   first_gv = netcount;
   tokens[0] = '\0';
   if (gndcount > 0) {
      gndptr = read_box("GND");
      gndnum = boxcount - 1;	// "GND" is stored in boxarr[gndnum]
      gnd_line = lcols + ucols + 1;  // order for GND cells
      boxarr[gndnum].count = gndcount;
   }
   if (vddcount > 0) {
      vddptr = read_box("Vdd");
      vddnum = boxcount - 1;
      vdd_line = lcols + ucols + (gnd_line ? 2 : 1); // order for Vdd cells
      boxarr[vddnum].count = vddcount;
      if (gndcount > 0 && vddptr->w != gndptr->w) {
	 fprintf(stderr, "Cells in \"Vdd.mag\" and \"GND.mag\" must have the "
	       "same width.\nPlease adjust the cells and try again.\n");
	 quit(1);
      }
   }

   // determine number of spacers to put in to make rows approximately the same
   // length
   maxcols = ucols + ABS(llength - ulength) / (vddnum ? boxarr[vddnum].w :
	 (gndnum ? boxarr[gndnum].w : SPACER_WIDTH)) + SPACERS;

   if (vddcount || gndcount) {
      if ((sprptr = read_box("spacer")) == NULL) {
	 fprintf(stderr,"Could not open \"spacer.mag\" for reading.\n");
	 quit(1);
      }
      sprnum = boxcount - 1;
      // store position of Vdd and GND ports in spacer
      if (vddcount)
         sprptr->vdd = vddptr->vdd;
      if (gndcount)
	 sprptr->gnd = gndptr->gnd;
      if ((vddcount && vddptr->w != sprptr->w) || (gndcount && gndptr->w !=
	       sprptr->w))
	 fprintf(stderr, "Cells in \"Vdd.mag\", \"GND.mag\", and \"spacer.mag\""
	       " must have the same width.\nPlease adjust the cells and try "
	       "again.\n");
      for (i = 0; i < vddcount + gndcount; i++) {
	 row = (ulength < llength ? UPPER : LOWER);
	 if (row)
	    ucols++;
	 else
	    lcols++;
	 create_cell(sprnum, row, (row) ? ucols : lcols, tokens);
	 if (row)
	    ulength += boxarr[sprnum].w;
	 else
	    llength += boxarr[sprnum].w;
	 if (lcols > maxcols || ucols > maxcols)
	    maxcols = (ucols > lcols ? ucols : lcols);
      }
      if (ucols < maxcols || lcols < maxcols) {
	 boxcount++;
	 boxarr = (boxptr)realloc(boxarr, sizeof(box) * boxcount);
	 sp2 = boxarr + boxcount - 1;
	 *sp2 = *sprptr;
	 sp2->name = malloc(sizeof(char) * 8);
	 strcpy(sp2->name, "spacer2");
	 sp2->w = sp2->numports = 0;
	 sp2->ports = NULL;
      }
   }
   else if (maxcols > ucols || lcols != ucols) {

      // check for predefined spacer cell.  If not present,
      // define two spacer boxes -- one with width 22 (width of nfet), one with
      // width 0 (to keep number of columns on top and bottom the same

      if ((sprptr = read_box("spacer")) == NULL) {
	 fprintf(stderr,"Could not open \"spacer.mag\" for reading.\n");
	 fprintf(stderr,"Creating generic spacer cells.\n");
         boxcount += 2;
         boxarr = (boxptr)realloc(boxarr, sizeof(box) * boxcount);
         sp1 = boxarr + boxcount - 2;
         sp2 = boxarr + boxcount - 1;
         sp1->name = malloc(sizeof(char) * 8);
         sp2->name = malloc(sizeof(char) * 8);
         strcpy(sp1->name, "spacer1");
         sp1->w = SPACER_WIDTH;
         strcpy(sp2->name, "spacer2");
         sp2->w = 0;
         sp1->ht = sp2->ht = sp1->instance = sp2->instance = sp1->count =
   	    sp2->count = sp1->numports = sp2->numports = 0;
         sp1->ports = sp2->ports = NULL;
         // position gnd and vdd at the same position as arbitrarily-chosen box
         sp1->vdd = sp2->vdd = boxarr[0].vdd;
         sp1->gnd = sp2->gnd = boxarr[0].gnd;
      }
      else {
	 sprnum = boxcount - 1;
	 boxcount++;
	 boxarr = (boxptr)realloc(boxarr, sizeof(box) * boxcount);
	 sp2 = boxarr + boxcount - 1;
	 *sp2 = *sprptr;
	 sp2->name = malloc(sizeof(char) * 8);
	 strcpy(sp2->name, "spacer2");
	 sp2->w = sp2->numports = 0;
	 sp2->ports = NULL;
      }
   }
   
   for (i = lcols; i < maxcols; i++)
      create_cell(boxcount - (llength > ulength ? 1 : 2), LOWER, i + 1,
	    tokens);
   for (i = ucols; i < maxcols; i++)
      create_cell(boxcount - (llength > ulength ? 2 : 1), UPPER, i + 1,
	    tokens);

   if (vdd_line || gnd_line)
      for (i = 0; i < sprcount; i++) {
	 (*(sprcells + i))->pos[GND] = boxarr[sprnum].gnd.ll;
	 (*(sprcells + i))->pos[VDD] = boxarr[sprnum].vdd.ll;
      }
   
   // deallocate memory

   for (i = 0; i < maxtokens; i++)
      if (*(tokens + i) != NULL)
	 free(*(tokens + i));
   free(tokens);

   return maxcols;
}

/*----------------------------------------------------------------------*/
/* Read in a flattened .spc file (only transistors)			*/
/*----------------------------------------------------------------------*/

int read_spice_fets(FILE *fi, char line_in[])
{
   int i, maxcols, *colptr, ncols = 0, pcols = 0;
   int j, numports;
   char **nets_in, model_name[100], cell_type;
   cellptr *source;
   int boxnum, nbox, pbox;

   fet_design = 1;
   nets_in = (char **)malloc(FETPORTS * sizeof(char *));
   for (i = 0; i < FETPORTS; i++)
      *(nets_in + i) = malloc(50 * sizeof(char));

   /* Assume standard SPICE port ordering; check for bulk connection, if any */

   while (fgets(line_in, 99, fi) != NULL) {
      if (line_in[0] != 'M') continue;

      remove_nl(line_in);
      sscanf(line_in, "%*s %49s %49s %49s %49s %99s", nets_in[IDRAIN],
	   nets_in[IGATE], nets_in[ISOURCE], nets_in[IBULK], model_name);
      boxnum = find_matching_box(model_name);
      if (boxnum < 0) quit(1);
      else if (boxnum == (boxcount - 1)) {
	 /* Consistency check:  are the ports in the right order? */
	 // fprintf(stderr, "To do: Need consistency check on MOSFET box %s!\n",
	 //		model_name);
      }

      switch(cell_type = model_name[0]) {
	  case 'n': case 'N':
		if (boxarr[boxnum].numports == FETPORTS - 1 &&
			strstr(nets_in[3], "GND") == NULL) {
		    fprintf(stderr, "Error: nfet layout does not define "
				"bulk/well connection but spice file connects"
				"\nbulk/well to %s, not GND.\n",
				cell_type, nets_in[3]);
		    quit(1);
		}
	        ncols++;
		colptr = &ncols;
		create_cell(boxnum, LOWER, *colptr, nets_in);
		source = lcells;
		nbox = boxnum;
		break;

	  case 'p': case 'P':
		if (boxarr[boxnum].numports == FETPORTS - 1 &&
			strstr(nets_in[3], "Vdd") == NULL) {
		    fprintf(stderr, "Error: pfet layout does not define "
				"bulk/well connection but spice file connects"
				"\nbulk/well to %s, not Vdd.\n",
				cell_type, nets_in[IBULK]);
		    quit(1);
		}
	        pcols++;
		colptr = &pcols;
		create_cell(boxnum, UPPER, *colptr, nets_in);
		source = ucells;
		pbox = boxnum;
		break;

	  default:
		fprintf(stderr, "Error: FET %s is not recognizable as N or P\n",
				model_name);
		continue;
      }

      // set the newcell's order
      (*(source + *colptr - 1))->order = ncols + pcols;
   }
   
   if (ncols > pcols) maxcols = ncols;
   else maxcols = pcols;

   maxcols += SPACERS;
   
   nets_in[0] = '\0';
   for (i = ncols; i < maxcols; i++)
      create_cell(nbox, LOWER, i + 1, nets_in);
   for (i = pcols; i < maxcols; i++)
      create_cell(pbox, UPPER, i + 1, nets_in);

   // clean up
   for (i = 0; i < FETPORTS; i++)
      free(*(nets_in + i));
   free(nets_in);
   
   first_gv = netcount;

   return (maxcols);
}

/*------------------------------------------------------------------*/

int read_spice(FILE *fi)
{
   int i, maxcol, filetype;
   int j, k;
   cellptr *source;
   char line_in[100];

   netcount = 0;
   networks = (netptr)malloc(sizeof(net));

   /* Force net numbers 0 and 1 to be GND and Vdd, respectively. */
   check_net("GND");
   check_net("Vdd");
   ucells = (cellptr *)malloc(sizeof(cellptr));
   lcells = (cellptr *)malloc(sizeof(cellptr));

   techname[0] = '\0';

   /* Look for subcircuits in the SPICE file */

   filetype = read_spice_subckts(fi, line_in);
   rewind(fi);

   switch(filetype) {
      case SUBCKTS:
      case GENCALLS:
	 maxcol = read_spice_general(fi, line_in);
	 break;
      case FETONLY:
	 maxcol = read_spice_fets(fi, line_in);
	 break;
      default:
	 fprintf(stderr, "Can't deal with this SPICE file\n");
	 fprintf(stderr, "Records should only be X or M, optionally with"
			" subcircuits.\n");			
	 maxcol = 0;
	 break;
   }
   return (maxcol);
}

/*----------------------------------------------------------------------*/
/* Determine how much blank space precedes the non-blank cells	 	*/
/* Return the number of units of blank space that can be eliminated.	*/
/* Need to find blank space on top and bottom separately due to		*/
/* presence of blank cells of width 0.					*/
/*----------------------------------------------------------------------*/

int calc_initial_blanks()
{
   int i, uspace, lspace;
   
   uspace = lspace = 0;
   for (i = 0; (i < cols) && ((*(lcells + i))->order == 0); i++)
      lspace += boxarr[(*(lcells + i))->boxnum].w;
   for (i = 0; (i < cols) && ((*(ucells + i))->order == 0); i++)
      uspace += boxarr[(*(ucells + i))->boxnum].w;
   
   return (uspace < lspace ? uspace : lspace);
}

/*--------------------------------------------------------------*/
/* Generate contact x positions for each cell			*/
/* and determine the endpoints of each network.			*/
/*--------------------------------------------------------------*/

int xpositions(int *misalignment)
{
   int i, j, k, x0, net, numports, source_idx, offset;
   int ll, ul, lr, ur;
   cellptr *source, thiscell, nextcell;
   boxptr thisbox;
   portinfoptr thisport;
 
   total_ports = 0;

   // convert all gnd/vdd cells back to spacers
   if (gnd_line || vdd_line) {
      for (i = 0; i < sprcount; i++)
	 if ((*(sprcells + i))->order == vdd_line ||
	       (*(sprcells + i))->order == gnd_line) {
	    (*(sprcells + i))->order = 0;
	    (*(sprcells + i))->boxnum = sprnum;
	    (*(sprcells + i))->net[GND] = GND;
	    (*(sprcells + i))->net[VDD] = VDD;
	 }
   } 
   
   // skip spacers in beginning.
   offset = -(calc_initial_blanks());

   for (j = 0; j < 2; j++) {
      x0 = offset;
      source = (j) ? ucells : lcells;
      for (i = 0; i < cols; i++) {
	 thiscell = *(source + i);
	 thiscell->overlap = 0;
	 thisbox = boxarr + thiscell->boxnum;
	 numports = thisbox->numports;

	 /* look for overlap source-to-source connection 	*/
	 /* (This to be re-done based on "Overlap" keywords	*/
	 /* in the constraints file for nfet/pfet)		*/
	 /*------

	 // This does not work!!!  Redo for new port idx method
	 if (fet_design && (thiscell->order > 0)) {
            if ((thiscell->orient == 0) && (i < cols - 1)) {
	       for (k = 0; (nextcell = *(source + i + k))->order == 0 &&
		     i + k < cols; k++);
               if ((k + i) < cols && (nextcell->orient == 1) &&
		     (thiscell->net[thiscell->ports[ISOURCE].idx] ==
			nextcell->net[nextcell->ports[ISOURCE].idx]))
	          thiscell->overlap = 1;
	    }
	 }

	 ------*/

	 // Initialize port x-positions, initialize begin and end of nets.
	 if (thiscell->order == 0) {
	    if (thisbox->w == 0)
	       continue;
	    if (gnd_line) {
	       thiscell->routed[GND] = 0;
	       thiscell->pos[GND].x = x0 + boxarr[gndnum].gnd.ll.x;
	    }
	    if (vdd_line) {
	       thiscell->routed[VDD] = 0;
	       thiscell->pos[VDD].x = x0 + boxarr[vddnum].vdd.ll.x;
	    }
	 }
	 else {
	    for (k = 0; k < numports; k++) {
	       thiscell->routed[k] = 0;
	       thiscell->pos[k].x = x0 + ((thiscell->orient) ?
		     (thisbox->w - (thisbox->ports[k].ll.x + m1_width)) :
		     thisbox->ports[k].ll.x);
	       thiscell->pos[k].y = thisbox->ports[k].ll.y;
	       net = thiscell->net[k];
	       if ((vdd_line || gnd_line) && net >= first_gv)
		  // reattach vdd/gnd ports to nets 0 and 1
		  net = thiscell->net[k] = ((networks + net)->name[0] == 'b');
	       if (net >= 0) {
	          if ((networks + net)->begin > thiscell->pos[k].x)
		      (networks + net)->begin = thiscell->pos[k].x;
	          if ((networks + net)->end < (thiscell->pos[k].x + m1_width))
		      (networks + net)->end = (thiscell->pos[k].x + m1_width);
	       }

	       /* Log the port */
	       total_ports++;
	       if (portorder == NULL)
		  portorder = (portinfoptr)malloc(sizeof(portinfo));
	       else
		  portorder = (portinfoptr)realloc(portorder,
				total_ports * sizeof(portinfo));
	       thisport = portorder + total_ports - 1;
	       thisport->source = source;
	       thisport->pos = i;
	       thisport->port = k;
	    }
         }
	 /* Overlap to be re-done based on contraints (see above) */
	 // Also needs to reflect new port index method!
	 /* x0 += (thiscell->overlap) ?
		(thisbox->ports[thisbox->ports[ISOURCE].idx].ll.x - 3) :
		thisbox->w; */
	
	if (i == 0) {
	    if (j == 0)
		ll = x0;
	    else
		ul = x0;
	}
	x0 += thisbox->w;
	if (i == (cols - 1)) {
	    if (j == 0)
		lr = x0;
	    else
		ur = x0;
	}
      }
   }

   /* Misalignment is the x difference between the first upper and lower */
   /* box plus the x difference between the last upper and lower box.	 */
   *misalignment = ABS(ll - ul) + ABS(lr - ur);

   /* Sort port portinfos by x-value */
   qsort(portorder, total_ports, sizeof(portinfo), picompare);

   return (-offset);
}

/*--------------------------------------------------------------*/
/* Randomly shuffle cell positions from current			*/
/*--------------------------------------------------------------*/

void scramble()
{
   int i;
   int p1, p2;
   cellptr tmpcell, *source1, *source2;

   for (i = 0; i < 20; i++) {
      p1 = random() % cols;
      p2 = random() % cols;
      source1 = (random() & 0x1) ? ucells : lcells;
      if (fet_design)
	 source2 = source1;	/* lower and upper rows cannot be permuted */
      else
         source2 = (random() & 0x1) ? ucells : lcells;

      tmpcell = *(source1 + p1);
      *(source1 + p1) = *(source2 + p2);
      *(source2 + p2) = tmpcell;
   }

   /* Choose a random orientation for each cell */

   for (i = 0; i < cols; i++) {
      (*(ucells + i))->orient = random() & 0x1;
      (*(lcells + i))->orient = random() & 0x1;
   }
}

/*--------------------------------------------------------------*/
/* Randomly select blocks of cells to swap and/or flip		*/
/*--------------------------------------------------------------*/

void place()
{
   int f, i, j, k, l, m, n, o, r, s;
   cellptr *source, tmpcell;
   int maxblock;
   char *placeh = placeinfo;

   maxblock = cols / 3;

   if (maxblock == 0) maxblock = 1;	 

   /* choose whether to move lower-cells, upper-cells, or both */
   i = (random() % 3) + 1;
   sprintf(placeh, "%s ", (i == 1) ? "(bottom row only)" :
		((i == 2) ? "(top row only)" : "(bottom and top)"));
   placeh = placeinfo + strlen(placeinfo);

   /* choose whether to flip when moving */

   f = random() % 2;
   
   /* for non-fet_designs choose whether to swap cells between row 1 	*/
   /* and row 2.  1 = switch horizontal, 2 = switch vertical, 3 = both	*/
   if (fet_design == 1)
      r = 1;
   else
      r = (random() % 3) + 1;
   
   /* Decide how many blocks to move from row 1 to row 2		*/
   s = random() % 2;

   /* choose block size to move */
   j = (random() % (maxblock + f)) + (1 - f); 

   /* choose offset of block 1 (start) */
   k = random() % (cols - j + (j != 0));

   /* on occasion, only one cell (or pair) will be flipped in orientation */
   /* but not swapped with another cell.				  */

   if (j == 0) {
      sprintf(placeh, "flip element at position %d ", k + 1);
      placeh = placeinfo + strlen(placeinfo);
      if (i & 1) (*(lcells + k))->orient ^= 0x1;
      if (i & 2) (*(ucells + k))->orient ^= 0x1;
      return;
   }

   /* choose offset of block 2 (start) such that block 2 does not overlap block 1 */

   for(;;) {
      if (cols == 2 )
      {
         l = ( k == 0 ? 1 : 0 );
	 break;
      }
      l = random() % (cols - j);
      if (((l + j) <= k) || (l >= (k + j))) break;
   }
   if (r != 2) {
      if (j == 1)
         sprintf(placeh, "swap %s position %d with %d ", ((f) ? "and flip"
		  : ""), k + 1, l + 1);
      else
         sprintf(placeh, "swap %s positions (%d to %d) with (%d to %d) ",
		((f) ? "and flip" : ""), k + 1, k + j, l + 1, l + j);
   }
   
   placeh = placeinfo + strlen(placeinfo);

   if (r != 1) {
      if (j == 1)
	 sprintf(placeh, "vertically swap cells in column %d ", k + 1);
      else
	 sprintf(placeh, "vertically swap cell in columns %d through %d ",
	       k + 1, k + j);
      if (s == 1) {
	 placeh = placeinfo + strlen(placeinfo);
	 if (j == 1)
	    sprintf(placeh, "and column %d ", l + 1);
	 else
	    sprintf(placeh, "and columns %d through %d ", l + 1, l + j);
      }
      if (f == 1 && r == 2) {
	 placeh = placeinfo + strlen(placeinfo);
	 sprintf(placeh, "and flip cells %s", (i == 1 ? "originally on bottom "
		  : (i == 2 ? "originally on top " : "")));
      }
   }

   placeh = placeinfo + strlen(placeinfo);
   
   if (r & 1) {
      source = (i & 1) ? lcells : ucells;
      for (m = 0; m < j; m++) {
	 n = (f & 1) ? (j - m - 1) : m;
	 tmpcell = *(source + k + m);
	 *(source + k + m) = *(source + l + n);
	 *(source + l + n) = tmpcell;
	 if (f & 1) {
	    (*(source + k + m))->orient ^= 0x1;
	    (*(source + l + n))->orient ^= 0x1;
	 }
      }
   }
   else if (f & 1) { // flip cells if not flipped above
      if (j > 1) {
	 source = (i & 1) ? lcells : ucells;
	 for (m = 0; m < 2; m++) {
	    n = (m ? k : l);
	    tmpcell = *(source + n);
	    *(source + n) = *(source + n + j - 1);
	    *(source + n + j - 1) = tmpcell;
	 }
      }
      for (m = 0; m < 2; m++) {
	 source = (m ? ucells : lcells);
	 for (n = 0; n < j; n++) {
	    (*(source + k + n))->orient ^= 0x1;
	    (*(source + l + n))->orient ^= 0x1;
	 }
      }
   }

   if (r & 2) {		// change rows (except in fet_designs)
      // always make first block change rows
      for (m = 0; m < j; m++) {
	 tmpcell = *(lcells + k + m);
	 *(lcells + k + m) = *(ucells + k + m);
	 *(ucells + k + m) = tmpcell;
      }
      if (s == 1) { // only make block 2 change row half of the time
	 for (m = 0; m < j; m++) {
	    tmpcell = *(lcells + l + m);
	    *(lcells + l + m) = *(ucells + l + m);
	    *(ucells + l + m) = tmpcell;
	 }
      }
   }
}

/*--------------------------------------------------------------*/

char netchar(int net)
{
   char tnet;
   
   if (net >= first_gv)
      return (networks + net)->name[0];

   if (net < 0) return ' ';
   tnet = (char)net + 'a';
   if (tnet > 'z') tnet = (char)net + 'A' - 26;
   return tnet;
}

/*--------------------------------------------------------------*/
/* Route by "right-hand rule" greedy algorithm.			*/
/* Initialize networks array and channels array.		*/
/* Returns: number of channels required for route, unless 	*/
/*  "magicout" is non-NULL, in which case it returns the status	*/
/*  of the magic route, either 0 for success or 1 for failure.	*/
/*  In both cases, -1 is returned on a blocked-port failure.	*/
/*--------------------------------------------------------------*/

int route(int *rlength, int *misalign, FILE *fileout, FILE *magicout,
		portinfoptr lastfailure)
{
   int i, j, k, l, num_channels, net, isfree, type, test, offset;
   int xpos, ocase, ptpflag, pidx, oidx, numports;
   int xmin, nlength, thisnet, pending, lastnet, lastpos, lastptpflag;
   chstructptr channels, thischannel;
   cellptr *source, thiscell, nextcell;
   portinfoptr cvals;
   netptr cnet, tnet;
   char *chlineout, this_direc, next_direc;
   boxptr tbox;
   constraintptr cons;
   int rightmost;
   int status = 0;

   channels = (chstructptr)malloc(sizeof(chstruct));
   num_channels = 0;
   *rlength = 0;
   lastnet = lastpos = pending = -1;

   for (i = 0; i < netcount; i++) {
      (networks + i)->metal = METAL1;   // horizontal nets are m1 by default
      (networks + i)->routed = 0;
      (networks + i)->begin = 1e6;
      (networks + i)->end = -1;
      (networks + i)->channel = -1;	// no channel assigned initially
   }

   /* Compute x positions of all ports according to orientation */
   /* and initialize begin and end of nets in network array.	*/

   offset = xpositions(misalign);
   
   /* Route ports connected to GND and Vdd (assign beginning and end	*/
   /* of lines and place lines in channels later.			*/
   if (vdd_line || gnd_line)
      gvroute(offset);

   /* Work from left to right; prune nets which have direct vertical or */
   /* point-to-point routes.						*/
   /* when top and bottom ports have same x; start with bottom		*/
   /* (arbitrary).						 	*/

   lastnet = lastpos = lastptpflag = 0;

   for (i = 0; i < total_ports; i++) {
      cvals = portorder + i; 
      thiscell = *(cvals->source + cvals->pos);
      if (thiscell->order == vdd_line || thiscell->order == gnd_line)
	 continue;

      pidx = cvals->port;
      net = thiscell->net[pidx];
      xpos = thiscell->pos[pidx].x;

      if (xpos == pending) pending = -1;

      /* This will reattach any nets that get shortened too much */
      /* by the algorithm directly below.			 */

      if (pending >= 0) {
	 if ((networks + net)->begin > xpos)
	    (networks + net)->begin = xpos;
	 if ((networks + net)->end < (xpos + m1_width))
	    (networks + net)->end = (xpos + m1_width);
      }
     
      if (net < 2) continue;	/* ports to Vdd or GND */

      j = 1;
      do {
         if (cvals->pos + j < cols)
            nextcell = *(cvals->source + cvals->pos + (j++));
	 else
	    nextcell = NULL;
      } while (nextcell != NULL && ((nextcell->order == 0) ||
	       nextcell->order == gnd_line || nextcell->order ==
	       vdd_line)); // skip blanks

      /* Nearest neighbor point-to-point connections (no channels	*/
      /* used). Situation is determined by both cell orientation and	*/
      /* which ports share the net;  e.g., most connections need to	*/
      /* be on facing sides of the cell.				*/

      if (nextcell != NULL) {
	 this_direc = (thiscell->orient ? 'w' : 'e');
	 next_direc = (nextcell->orient ? 'e' : 'w');
	 ptpflag = check_for_ptp(thiscell, nextcell, this_direc, next_direc,
	       pidx, &oidx, cvals->source == ucells);

	 /* Shorten the network by the length of the point-to-point	*/
	 /* connection, if applicable.					*/

	 if (ptpflag) {
	    int xnpos = nextcell->pos[oidx].x;
	    if (xpos <= (networks + net)->begin) {
	       if (!(fet_design && lastnet == net && lastpos == xpos &&
			!lastptpflag))
	          (networks + net)->begin = xnpos;
	       if (lastnet != net || lastpos != xpos || (fet_design &&
			lastnet == net && lastpos == xpos && lastptpflag))
		  thiscell->routed[pidx] = 1;
	       pending = xnpos;
	    }
            else if (xnpos >= ((networks + net)->end - m1_width)) {
	       (networks + net)->end = xpos + m1_width;
	       nextcell->routed[oidx] = 1;
	       pending = xnpos;
	    }
	    /* if both connections are in the middle of the net, */
	    /* mark the first (arbitrarily) as being routed.	 */
	    else
	       thiscell->routed[pidx] = 1;
	 }
      }
      lastnet = net;
      lastpos = xpos;
      lastptpflag = ptpflag;
   }

   // Make sure ports on bottom can route north and ports on top can route south
   if (!fet_design) {
      for (i = 0; i < ROWS; i++) {
	 source = (i ? ucells : lcells);
	 for (j = 0; j < cols; j++) {
	    thiscell = *(source + j);
	    if (thiscell->order == 0) // skip blank cells
	       continue;
	    tbox = boxarr + thiscell->boxnum;
	    for (k = 0; k < tbox->numports; k++) {
	       if (thiscell->routed[k] == 1 || thiscell->net[k] < 2 ||
		     (((networks + thiscell->net[k])->end -
		      (networks + thiscell->net[k])->begin) < PORT_SEP &&
		      !matching_v(thiscell->pos[k].x, thiscell->net[k],
			 (i ? lcells : ucells), j)))
		  continue;
	       if ((i == LOWER && tbox->ports[k].north == 0) ||
		     (i == UPPER && tbox->ports[k].south == 0)) {
		  boxptr lastbox = NULL;
		  if (lastfailure->pos >= 0)
		     lastbox = boxarr + (*(lastfailure->source +
			   lastfailure->pos))->boxnum;
		  if (lastfailure->pos < 0 || lastfailure->port != k ||
				lastbox->name != tbox->name) {
		     fprintf(stderr, "Port \"%s\" of cell \"%s\" cannot be "
			   "routed %s.\n", tbox->ports[k].name, tbox->name,
			   (tbox->ports[k].north == 0 ? (tbox->ports[k].south
			    == 0 ? "north or south" : "north") : "south"));
		     lastfailure->source = source;
		     lastfailure->pos = j;
		     lastfailure->port = k;
		  }
		  return -1;	/* Blocked-port failure */
	       }
	    }
	 }
      }
   }

   /* route each net in order and compute total net length on the fly	*/
   /* Initialize "routed" and "channel" of networks.			*/
   /* Initialize channels.						*/

   rightmost = 0;
   for (i = 2; i < netcount; i++) {
      cnet = networks + i;

      if (cnet->end > rightmost) rightmost = cnet->end;

      /* Parts of point-to-point connections may be ignored */
      if (cnet->routed) continue;

      /* Cull out routes of size zero */
      if ((nlength = (cnet->end - cnet->begin)) < PORT_SEP) {
	 cnet->routed = 1;
	 continue;
      }

      *rlength += nlength;

      for (k = 0; k < num_channels; k++) {
	 thischannel = (channels + k);
	 isfree = 1;
	 for (l = 0; l < thischannel->num_nets; l++) {
	    tnet = networks + *(thischannel->nets + l);
	    if ((tnet->begin < cnet->end) && (tnet->end > cnet->begin)) {
	       isfree = 0; 
	       break;
	    }
	 }
	 if (isfree) break;
      }
      cnet->channel = k;
      cnet->routed = 1;
      if (k == num_channels) {  /* make a new channel and fill it */
	 num_channels++;
   	 channels = (chstructptr)realloc(channels, num_channels *
	       sizeof(chstruct));
	 thischannel = (channels + k);
	 thischannel->num_nets = 1;
	 thischannel->nets = (int *)malloc(sizeof(int));
	 thischannel->lucount = 0;
	 *(thischannel->nets) = i;
      }
      else {	/* add net to existing channel */
	 thischannel->num_nets++;
	 thischannel->nets = (int *)realloc(thischannel->nets,
		thischannel->num_nets * sizeof(int));
	 *(thischannel->nets + thischannel->num_nets - 1) = i;
      }
   }

   /* determine number of lower-connections vs. upper-connections on each */
   /* channel */

   for (i = 0; i < cols; i++) {
      for (j = 0; j < ROWS; j++) {
	 source = (j) ? ucells : lcells;
	 thiscell = *(source + i);
	 if (thiscell->order == 0)
	    continue;
	 numports = boxarr[thiscell->boxnum].numports;
	 for (k = 0; k < numports; k++) {
	    if ((networks + thiscell->net[k])->channel >= 0)
	          (channels + (networks + thiscell->net[k])->channel)->lucount
		     +=	((j) ? 1 : -1);
         }
      }
   }

   /* re-order the channels to minimize vertical route distances */

   qsort(channels, num_channels, sizeof(chstruct), chcompare);

   /* Generate a route-length cost for constraints */

   k = 0;
   for (cons = constraints; cons != NULL; cons = cons->next) {
      k++;
      cnet = networks + cons->thisnet;
      nlength = 0;
      switch(cons->type) {
	 case LEFT:
	    nlength = cnet->begin;
	    break;
	 case RIGHT:
	    nlength = rightmost - cnet->end;
	    break;
	 case TOP:
	    /* No action, yet */
	    break;
	 case BOTTOM:
	    /* No action, yet */
	    break;
	 case TOGETHER:
	    tnet = networks + cons->thatnet;
	    if (cnet->begin > tnet->end)
	       nlength = cnet->begin - tnet->end;
	    else if (cnet->end < tnet->begin)
	       nlength = tnet->begin - cnet->end;
	    break;
      }
      nlength -= cons->allow;
      if (nlength > 0)
	 *rlength += cons->weight * nlength;
   }

   /* Print results */

   if (fileout != NULL)
      draw_layout(channels, num_channels, fileout);

   /* Generate .mag layout file */

   if (magicout != NULL)
      status = gen_magic(channels, num_channels, magicout);

   /* free channel memory */

   for (i = 0; i < num_channels; i++)
      free((channels + i)->nets);
   free(channels);

   if (magicout != NULL)
      return status;
   else
      return num_channels;
}

/*------------------------------------------------------------------*/

write_nets(FILE *fileout)
{
   int i;

   if (fileout == stderr || fileout == NULL) {
      fprintf(fileout, "Network names:\n");
      for (i = 0; i < first_gv; i++) 
         fprintf(fileout, "%c: %s\n", netchar(i), (networks + i)->name);
      fprintf(fileout, "\n");
   }
   else {
      int yoff = 72 + 16 * (first_gv / 2);
      fprintf(fileout, "10 %d moveto (Network names:) show\n", yoff);
      for (i = 0; i < first_gv; i++) {
	 if (!(i & 1)) yoff -= 16;
         fprintf(fileout, "%d %d moveto (%c:) show ( %s) show\n",
		(i & 1) ? 250 : 20,
		yoff, netchar(i), (networks + i)->name);
      }
   }
}

/*--------------------------------------------------------------*/
/* Check if a solution is unique or not				*/
/*--------------------------------------------------------------*/

int is_unique(infoptr keep, int thisidx, int searchend)
{
   int j, k;
 
   for (k = 0; k < searchend; k++) {
      if ((keep[thisidx].cost == keep[k].cost) &&
	  (keep[thisidx].rlength == keep[k].rlength) &&
	  (keep[thisidx].channels == keep[k].channels)) {
	 for (j = 0; j < cols; j++) {
	    if ((keep[thisidx].upper[j].cell != keep[k].upper[j].cell) ||
		(keep[thisidx].upper[j].orient != keep[k].upper[j].orient) ||
		(keep[thisidx].lower[j].cell != keep[k].lower[j].cell) ||
		(keep[thisidx].lower[j].orient != keep[k].lower[j].orient))
	       break;
	 }
	 if (j == cols) return 0;	/* not unique */
      }
   }
   return 1;	/* is unique */
}

/*--------------------------------------------------------------*/
/* Sort the solution population according to cost with minimum	*/
/* cost at the top.  Prune the list so that all entries	are	*/
/* unique.  Return the number of unique solutions.		*/
/*--------------------------------------------------------------*/

int sort_solutions(infoptr keep)
{
   int i, unique = 1;

   qsort(keep, POPULATION, sizeof(info), infocompare);

   /* Retain only unique individuals.			*/
   /* Original behavior:  return POPULATION and ignore	*/
   /* the rest of this subroutine.			*/

   for (i = 1; i < POPULATION; i++) {
      if (is_unique(keep, i, unique)) {
	 if (i != unique)
	    icopyfwd(&keep[i], &keep[unique]);
	 unique++;
      }
   }
   return unique;
}

/*--------------------------------------------------------------*/
/* Redistribute the population of solutions, giving greater	*/
/* weight (probability) to the best solutions 			*/
/*--------------------------------------------------------------*/

redistribute(infoptr keep)
{
   float prob[POPULATION], ptotal;
   int copies[POPULATION], ctotal;
   int i, j, k, unique;

   /* sort the population and cull out redundant solutions */

   unique = sort_solutions(keep);

   /* Generate exponential curve weighted towards the best solution */

   ptotal = 0.0;
   for (i = 0; i < unique; i++) {
      prob[i] = (float)(keep[unique - 1].cost / keep[i].cost);
      prob[i] = exp((double)prob[i]);
      ptotal += prob[i];
   }
   ptotal = POPULATION / ptotal;

   /* copies[] is the number of copies we will make of each solution. */

   ctotal = 0;
   for (i = 0; i < unique; i++) {
      copies[i] = prob[i] * ptotal;
      ctotal += copies[i];
   }
   for (i = unique; i < POPULATION; i++)
      copies[i] = 0;

   /* Adjust for round-off errors */

   for (i = 0; ctotal < POPULATION; i++) {
      copies[i]++;
      ctotal++;
   }
   for (i = unique - 1; ctotal > POPULATION; i--) {
      if (copies[i] > 0) {
         copies[i]--;
         ctotal++;
      }
   }

   /* Repopulate the solutions. */

   j = POPULATION - 1;
   for (i = POPULATION - 1; i >= 0; i--) {
      for (k = 0; k < copies[i]; k++) {
	 icopyfwd(&keep[i], &keep[j]);
	 j--;
      }
   }
}

/*--------------------------------------------------------------*/
/* Cost function						*/
/*--------------------------------------------------------------*/

int compute_cost(int channels, int rlength, int misalignment)
{
   int cost;

   /* First, compute cost for each channel in excess of the	*/
   /* target number of channels.				*/

   cost = channels - ch_minimum;
   if (cost < 0) cost = 0;

   /* Weight this cost by the penalty per column */

   cost *= (ch_penalty * cols);

   /* Add the cost of the horizontal routes */

   cost += rlength;

   /* Add the cost of row misalignment */

   cost += (align_penalty * misalignment);

   return cost;
}

/*--------------------------------------------------------------*/
/* Main program							*/
/*--------------------------------------------------------------*/

int main(int argc, char *argv[])
{
   char c, filename[100], line_in[100];
   char *dotptr;
   FILE *filein, *fileout, *layout;
   info *keep;
   int i, j, channels, rlength, startpos, misalign;
   int failcount, popcount;
   int cost, minidx;
   int fd = fileno(stdin);
   unsigned long iter, best_iter;
   portinfo failure;
   constraintptr cons, newcons;
   int status;
   int width;
   float scale;

   // Check for global "Definitions" constraints file.
   // This sets values for m1_width, etc., which will be required when
   // evaluating the layouts.
   // NULL may be passed to read_definitions(), in which case layout
   // properties are set to default (compiled-in) values.
   
   filein = fopen("Definitions", "r");
   read_definitions(filein);		/* filein may be NULL; sets defaults */

   if (filein == NULL)
      fprintf(stderr, "No global definitions (\"Definitions\") in this directory.\n");
   else
      fclose(filein);

   sprintf(filename, argv[1]);
   if ((dotptr = strrchr(filename, '.')) == NULL)
      dotptr = filename + strlen(filename);
   else if (strrchr(filename, '/') > dotptr)
      dotptr = filename + strlen(filename);
   sprintf(dotptr, ".fspc");

   if ((filein = fopen(filename, "r")) == NULL) {
      dotptr = strrchr(filename, '.');
      sprintf(dotptr, ".spc");
      if ((filein = fopen(filename, "r")) == NULL) {
	 fprintf(stderr, "Can't open file \"%s\" for reading\n",
	 	filename);
	 fprintf(stderr, "Usage:  genrouter filename[.spc]\n");
      	 quit(1);
      }
   }
   fprintf(stderr, "Reading spice file.\n");
   cols = read_spice(filein);
   fclose(filein);
   
   // check for circuit ".cons" constraints file.

   constraints = (constraintptr) NULL;
   sprintf(dotptr, ".cons");
   if ((filein = fopen(filename, "r")) == NULL)
      fprintf(stderr, "No local contraints (\"%s\") for this layout.\n", filename);
   else {
      read_constraints(filein);
      fclose(filein);
   }

   // start random number generation from specific point if requested,
   // otherwise create new layout.
   if (argc > 2 && strrchr(argv[2], 's') != NULL) {
      dotptr = strrchr(filename, '.');
      sprintf(dotptr, ".ps");
      if ((filein = fopen(filename, "r")) == NULL) {
	 fprintf(stderr, "Could not find \"%s\" to reproduce old layout.\n"
	       "Continuing with new layout.\n", filename);
	 startpos = setrandom(0);
      }
      else
	 for (i = 0; i < 2; i++) 		/* skip two lines of PostScript */
	    fgets(line_in, 99, filein);			
         fscanf(filein, "%*s %*c %d", &startpos);
         startpos = setrandom(startpos);
	 fclose(filein);
   }
   else
      startpos = setrandom(argc > 2 ? atoi(argv[2]) : 0);
   fprintf(stderr, "Random number generation began with %d.\n", startpos);
   iter = 0;

   /* Write out the list of nets and their one-letter designations */
   write_nets(stderr);

   best.upper = (cellinfoptr)malloc(cols * sizeof(cellinfo));
   best.lower = (cellinfoptr)malloc(cols * sizeof(cellinfo));

   keep = (infoptr)malloc(POPULATION * sizeof(info));

   failure.source = lcells;
   failure.pos = -1;
   failure.port = -1;

   portorder = NULL;
   total_ports = 0;

   for (j = 0; j < POPULATION; j++) {
      keep[j].upper = (cellinfoptr)malloc(cols * sizeof(cellinfo));
      keep[j].lower = (cellinfoptr)malloc(cols * sizeof(cellinfo));

      for (i = 0; (channels = route(&rlength, &misalign, NULL, NULL,
			&failure)) == -1; i++) {
         scramble();
         if (i == 1000) {
	    fprintf(stderr, "Tried %d layouts without finding a viable one.  \n"
	       "Please check subcells for blocked ports and try again.\n\n", i);
	    quit (1);
         }
      }
      ccopyfwd(ucells, lcells, &keep[j]);
      keep[j].rlength = rlength;
      keep[j].misalign = misalign;
      keep[j].channels = channels;
      keep[j].cost = compute_cost(channels, rlength, misalign);
   }

   /* Find the minimum cost */
   minidx = 0;
   for (j = 1; j < POPULATION; j++) {
      if (keep[j].cost < keep[minidx].cost) {
	 minidx = j;
      }
   }

   /* Copy the solution with the minimum cost to the "best" record */
   icopyfwd(&keep[minidx], &best);
   ccopyback(&best, ucells, lcells);

   /* Open graphics window and print the best solution */
   width = open_window();
   scale = (float)468 / (float)width;

   /* Print the best solution */
   route(&rlength, &misalign, stderr, NULL, &failure);

   fprintf(stderr, "Initial best route.  Channels: %d; Route length: %d;  Cost: %d\n",
		best.channels, best.rlength, best.cost);

   fprintf(stderr, "%s\n\n",
	"Type 'w' to end iterations and (over)write result, 's' backup any\n"
       	"previously produced layout and then end iterations and write result,\n"
       	"'p' to pause, 'q' to abandon.");
   fprintf(stderr, "Hit <return> to start:");
   c = fgetc(stdin);
   fprintf(stderr, "\n\n");
   setnodelay(fd);

   for(;;) {
      while (read(fd, &c, 1) <= 0) {
	 for (i = 0; i < POPULATION; i++) {

	    /* Based on MUTATION_RATE, either mutate the solution or leave it	*/
	    /* alone and go on to the next one.					*/

	    if ((random() % 100) < MUTATION_RATE) {

	       /* get the solution */
	       ccopyback(&keep[i], ucells, lcells);

	       /* randomly mutate the solution */
               place();

	       /* place & route;  repeat until the solution is routable */
	       failcount = popcount = 0;
	       while ((channels = route(&rlength, &misalign, NULL,
				NULL, &failure)) == -1) {
	          failcount++;
		  if (failcount == 1000) {
		     popcount++;
	             fprintf(stderr, "\nTried %d layouts without finding a viable"
		        		" one.\n", failcount);

		     /* overwrite this solution with another one */
		     icopyfwd(&keep[(i + popcount) % POPULATION], &keep[i]);

		     failcount = 0;
		     if (popcount == POPULATION) {
			fprintf(stderr, "\nMajor snag: This circuit is not"
					" routable.  Quitting\n");
			quit(1);
		     }
		  }
		  ccopyback(&keep[i], ucells, lcells);	/* revert to original */
		  place();				/* try another mutation */
	       }
   	       cost = compute_cost(channels, rlength, misalign);

	       if (cost < best.cost) {
	          ccopyfwd(ucells, lcells, &best);	/* Save it! */
	          best.cost = cost;
	          best.channels = channels;
	          best.rlength = rlength;
	          best.misalign = misalign;

	          best_iter = iter;
	          fprintf(stderr, "\nStep %d: ", iter);
	          fprintf(stderr, "   %s\n", placeinfo);
	          fprintf(stderr, "   Channels: %d; Route length: %d; Cost: %d\n\n",
					channels, rlength, cost);
		  /* Print solution to stderr */
                  channels = route(&rlength, &misalign, stderr, NULL, &failure);
	       }

	       ccopyfwd(ucells, lcells, &keep[i]);	/* write back the solution */
	       keep[i].cost = cost;
	       keep[i].rlength = rlength;
	       keep[i].misalign = misalign;
	       keep[i].channels = channels;
	    }
	 }
	 redistribute(keep);	/* Redistribute the population */
	 fprintf(stderr, ".");	/* Mark the cycle as completed */
	 iter++;
      }
      if (c == 'p') {
	 fprintf(stderr, "\nHit <return> to continue. . .");
	 setnormal(fd);
	 c = fgetc(stdin);
	 setnodelay(fd);
	 fprintf(stderr, "\n");
      }
      else if (c == ' ') {	/* Refresh graphics */
	 ccopyback(&best, ucells, lcells);
         route(&rlength, &misalign, stderr, NULL, &failure);
	 ccopyfwd(ucells, lcells, &best);
      }
      else break;
   }
   tcflush(fd, TCIFLUSH);
   setnormal(fd);
   fprintf(stderr, "\n");

   if (c == 's') {
      char *bkpname;
      char *bkpdotptr;
      sprintf(dotptr, ".ps");
      bkpname = (char *)malloc(strlen(filename) + 6);
      strcpy(bkpname, filename);
      bkpdotptr = strrchr(bkpname, '.');
      for (i = 0; i < 1000; i++) {
         sprintf(bkpdotptr, "_%d.ps", i);
         if ((fileout = fopen(bkpname, "r")) == NULL)
            break;
         fclose(fileout);
      }
      if ((filein = fopen(filename, "r")) != NULL) {
	 fprintf(stderr, "\nMoving %s to %s.\n", filename, bkpname);
	 // copy .ps file
	 fclose(filein);
	 rename(filename, bkpname);
      }
      sprintf(dotptr, ".mag");
      if ((filein = fopen(filename, "r")) != NULL) {
	 sprintf(bkpdotptr, "_%d.mag", i);
	 fprintf(stderr, "\nMoving %s to %s.\n", filename, bkpname);
	 // copy .ps file
	 fclose(filein);
	 rename(filename, bkpname);
      }
      free(bkpname);
   }
   if (c != 'q') {
      ccopyback(&best, ucells, lcells);
      sprintf(dotptr, ".ps");
      fprintf(stderr, "\nWriting results to file \"%s\"\n", filename);
      if ((fileout = fopen(filename, "w")) == NULL) {
         fprintf(stderr, "Can't open file \"%s\" for writing\n", filename);
	 fprintf(stderr, "Results are drawn on the graphics window.\n");
	 fileout = stderr;
      }
      sprintf(dotptr, ".mag");
      if ((layout = fopen(filename, "w")) == NULL)
         fprintf(stderr, "Can't open file \"%s\" for writing layout\n",
		filename);

      if (fileout != stderr) {
	 fprintf(fileout, "%%!PostScript output from genrouter-1.9\n");
	 fprintf(fileout, "%%%%BoundingBox: 0 0 612 792\n");
      }

      fprintf(fileout, "%%Random_number_startpos = %d\n", startpos);
      fprintf(fileout, "%%Layout number %d\n%%Cost = %d\n", best_iter, best.cost);

      if (fileout != stderr) {
	 fprintf(fileout, "gsave\n");
	 fprintf(fileout, "1 setlinecap 0 setlinejoin 6 setmiterlimit 0 setgray\n");
	 fprintf(fileout, "72 72 translate %3.2f %3.2f scale\n", scale, scale);
	 fprintf(fileout, "/Helvetica findfont 14 scalefont setfont\n");
      }
      write_nets(fileout);

      /* save route results to text file and create layout in .mag file */
      status = route(&rlength, &misalign, fileout, layout, &failure);

//    status = route(&rlength, &misalign, NULL, layout, &failure);

      if (status == -1) {
	 fprintf(stderr, "Last route is not viable.  Ports are blocked.\n");
      }

//    else if (status == 1) {
//	 int unique;
//	 /* Go through the list of solutions until one is routable */
//	 fprintf(stderr, "Best solution is not viable.  Checking other solutions.\n");
//	 unique = sort_solutions(keep);
//	 for (i = 0; i < unique; i++) {
//	    rewind(layout);
//	    ccopyback(&keep[i], ucells, lcells);
//	    status = route(&rlength, &misalign, NULL, layout, &failure);
//	    if (status == 0) break;
//	 }
//	 if (i == unique) {
//	    fprintf(stderr, "No good solution was found.  Printing best solution.\n");
//	    fprintf(stderr, "Layout will require hand-editing.\n");
//	    rewind(layout);
//	    ccopyback(&best, ucells, lcells);
//	    /* ignore return value */
//	    route(&rlength, &misalign, NULL, layout, &failure);
//	 }
//    }
//
//    /* Finally, save the route information */
//    status = route(&rlength, &misalign, fileout, NULL, &failure);
      
      // report any unmarked input/output pins in final magic layout
      for (i = 0; i < netcount; i++) {
	 if ((networks + i)->name != NULL && strstr((networks + i)->name,
	       "int") == NULL && strstr((networks + i)->name, "ext") == NULL &&
	       strstr((networks + i)->name, "GND") == NULL &&
	       strstr((networks + i)->name, "Vdd") == NULL && i < first_gv) {
	    fprintf(stderr, "Net \"%s\" not marked in magic file.\n",
		  (networks + i)->name);
	    if (fileout != stderr)
	       fprintf(fileout, "%%Net \"%s\" not marked in magic file.\n",
		     (networks + i)->name);
	 }
      }

      if (fileout != stderr) {
	 fprintf(fileout, "grestore showpage\n");
	 fclose(fileout);
      }
      if (layout != NULL) fclose(layout);
   }

   /* Cleanup: free memory. */

   for (i = 0; i < netcount; i++)
      if ((networks + i)->name != NULL)
         free((networks + i)->name);
   free(networks);

   cons = constraints;
   while (cons != NULL) {
      newcons = cons->next;
      free(cons);
      cons = newcons;
   }

   free(sprcells);
   free(portorder);

   for (i = 0; i < cols; i++) {
      free((*(ucells + i))->routed);
      free((*(ucells + i))->net);
      free((*(ucells + i))->pos);
      free((*(lcells + i))->routed);
      free((*(lcells + i))->net);
      free((*(lcells + i))->pos);
      free(*(ucells + i));
      free(*(lcells + i));
   }

   for (i = 0; i < boxcount; i++) {
      if ((boxarr + i)->name != NULL)
	 free((boxarr + i)->name);
      for (j = 0; j < (boxarr + i)->numports; j++)
	 free(((boxarr + i)->ports + j)->name);
      if ((boxarr + i)->numports != 0) {
	 free((boxarr + i)->ports);
      }
   }
   free(boxarr);

   for (i = 0; i < POPULATION; i++) {
      free(keep[i].upper);
      free(keep[i].lower);
   }
   free(best.upper);
   free(best.lower);
   
   quit(0);
}

/*------------------------------------------------------------------*/
