/*--------------------------------------------------------------*/
/* gnd_vdd.c							*/
/*	Functions used to manipulate GND, Vdd, and spacer	*/
/* 	cells used in some subcircuit designs.			*/
/*								*/
/* Copyright 2001 - Jennifer Roberts				*/	
/*--------------------------------------------------------------*/

#include "defines.h"

int vdd_line = 0;
int gnd_line = 0;
int vddnum = 0, gndnum = 0, sprnum = 0;	// boxnums of vdd, gnd, and spacer boxes
cellptr *sprcells;		// contains pointers to all spacer cells
int sprcount;
int first_gv;			// num of first "net" in networks array.

extern int m1_width;
extern int m1_space;
extern int m2_width;
extern int m2_space;
extern int via_width;

extern int total_ports;
extern portinfoptr portorder;

/*--------------------------------------------------------------*/
/* Return 1 if given vdd/gnd port can be connected to an	*/
/* adjacent vdd/gnd cell.  If so, change order and boxnum of	*/
/* cell to reflect that it is to be used as a vdd/gnd cell,	*/
/* and set port's routed to indicate type of point-to-point.	*/
/*--------------------------------------------------------------*/

int gv_ptp(int src_ln, int bxnum, cellptr *source, int colnum, int portnum)
{
   int i, routable, row, ly, uy, py;
   cellptr ncell, tcell = (*(source + colnum));
   boxptr tbox = boxarr + tcell->boxnum;

   py = boxarr[tcell->boxnum].ports[portnum].ll.y;
   ly = ((src_ln == vdd_line) ? boxarr[vddnum].vdd.ll.y :
	 boxarr[gndnum].gnd.ll.y);
   uy = ((src_ln == vdd_line) ? boxarr[vddnum].vdd.ur.y :
	 boxarr[gndnum].gnd.ur.y);
   row = (source == ucells ? UPPER : LOWER);
   routable = (tcell->orient ? tbox->ports[portnum].west :
	 tbox->ports[portnum].east);
   if (routable && !(row == LOWER && routable == IF_ON_TOP) &&
	 !(row == UPPER && routable == IF_ON_BOTTOM)) {
      // check for spacer/vdd/gnd cell to the right
      for (i = colnum + 1; i < cols; i++)
         // skip over cells of 0 width
         if (boxarr[(*(source + i))->boxnum].w != 0)
	    break;
      if (i != cols) {
         ncell = *(source + i);
         if ((ncell->order == 0 || ncell->order == src_ln) &&
	       py < uy && py >= ly) {
            ncell->order = src_ln;
            ncell->boxnum = bxnum;
            tcell->routed[portnum] = PTP_TO_RIGHT;
	    return 1;
	 }
      }
   }
   
   routable = (tcell->orient ? tbox->ports[portnum].east :
	 tbox->ports[portnum].west);
   if (routable && !(row == LOWER && routable == IF_ON_TOP) &&
	 !(row == UPPER && routable == IF_ON_BOTTOM)) {
      // check for spacer/vdd/gnd cell to the left
      for (i = colnum - 1; i >= 0; i--)
	 // skip over cells of 0 width
         if (boxarr[(*(source + i))->boxnum].w != 0)
	    break;
      if (i >= 0) {
	 ncell = *(source + i);
	 if ((ncell->order == 0 || ncell->order == src_ln) &&
	       py < uy && py >= ly) {
            ncell->order = src_ln;
            ncell->boxnum = bxnum;
            tcell->routed[portnum] = PTP_TO_LEFT;
	    return 1;
	 }
      }
   }
   
   return 0;
}

/*--------------------------------------------------------------*/
/* Turn closest spacer/vdd/gnd cell into vdd/gnd cell.		*/
/* Return 1 if successful.					*/
/*--------------------------------------------------------------*/

cellptr find_closest_gv(int src_ln, cellptr tcell, int portnum)
{
   cellptr closest_cell, *source;
   int i, k, distance, portx, xpos = tcell->pos[portnum].x;
   
   distance = 1000000;
   for (k = 0; k < ROWS; k++) {
      source = (k ? ucells : lcells);
      // find closest vdd/gnd/spacer cell
      for (i = 0; i < cols; i++) {
	 if (((*(source + i))->order != 0 && (*(source + i))->order !=
	       src_ln) || boxarr[(*(source + i))->boxnum].w == 0)
	    continue;
	 portx = (src_ln == vdd_line ? (*(source + i))->pos[VDD].x :
	       (*(source + i))->pos[GND].x);
         if ((portx <= xpos) && abs(portx - xpos) < distance) {
	    distance = abs(portx - xpos);
	    closest_cell = (*(source + i));
         }
         if (portx + m1_width >= xpos) {
	    if (abs(portx - xpos) < distance) {
	       distance = abs(portx - xpos);
	       closest_cell = (*(source + i));
	    }
	    break;
         }
      }
   }
   if (distance == 1000000)
      return NULL; // failed to find spacer cell.

   return closest_cell;
}

/*--------------------------------------------------------------*/

netptr create_gvnet(int type, int portx)
{
   netptr newnet;
   
   netcount++;
   networks = (netptr)realloc(networks, sizeof(net) * netcount);
   newnet = networks + netcount - 1;

   newnet->name = malloc(sizeof(char) * 2);
   newnet->name[0] = 'a' + type;
   newnet->name[1] = '\0';
   newnet->begin = portx;
   newnet->end = portx + m1_width;
   newnet->channel = -1;
   newnet->routed = 0;
   
   return newnet;
}

/*--------------------------------------------------------------*/

int vddcellcmp(cellptr *a, cellptr *b)
{
   if ((*a)->pos[VDD].x == (*b)->pos[VDD].x)
      return 0;
   else return ((*a)->pos[VDD].x < (*b)->pos[VDD].x ? -1 : 1);
}

/*--------------------------------------------------------------*/

int gndcellcmp(cellptr *a, cellptr *b)
{
   if ((*a)->pos[GND].x == (*b)->pos[GND].x)
      return 0;
   else return ((*a)->pos[GND].x < (*b)->pos[GND].x ? -1 : 1);
}

/*--------------------------------------------------------------*/

void convert_to_gv(cellptr gvcell, int srcline, int bxnum)
{
   netptr tnet = (networks + netcount - 1);
   int idx = (srcline == gnd_line ? GND : VDD);
   int xpos = gvcell->pos[idx].x;  // pos of GND/Vdd port
   
   if (tnet->routed == 1)
      return;  // line is already attached to another gnd/vdd cell
   
   gvcell->order = srcline;
   gvcell->boxnum = bxnum;
   gvcell->net[idx] = netcount - 1;
   
   // extend net to attach it to gvcell
   if (xpos < tnet->begin)
      tnet->begin = xpos;
   else if (xpos + m1_width > tnet->end)
      tnet->end = xpos + m1_width;
   tnet->routed = 1;
}

/*--------------------------------------------------------------*/

cellptr find_intermediate_cell(int x1, int x2, int line)
{
   int i;
   
   for (i = 0; i < sprcount; i++)
      if ((*(sprcells + i))->pos[line].x > (x1 - PORT_SEP) &&
	    (*(sprcells + i))->pos[line].x < (x2 + PORT_SEP))
	 return (*(sprcells + i));
   
   return NULL;
}

/*--------------------------------------------------------------*/
/* Route ports connected to GND or Vdd.				*/
/* Search for point-to-points and determine beginning and end	*/
/* points of lines to be routed in channels.  Convert the	*/
/* necessary spacer cells to gnd/vdd cells.			*/
/*--------------------------------------------------------------*/

void gvroute(int init_offset)
{
   int i, j, pidx, xpos, line, src_ln, bxnum, endx, final_offset;
   int newlength, tolastp, tolastgv, backlength, lastpx;
   cellptr tcell, gvcell, begin_gv, lastgv, *source;
   netptr pending;
   portinfoptr cvals;

   // clear gnd/vdd connections from last layout
   for (i = first_gv; i < netcount; i++)
      if ((networks + i)->name != NULL) {
	 free((networks + i)->name);
	 (networks + i)->name = NULL;
      }
   netcount = first_gv;
   networks = realloc(networks, sizeof(net) * netcount);
   
   // find offset of layout without gnd/vdd cells

   // decide whether to route GND or Vdd first (route one with most cells)
   line = (!gnd_line || boxarr[vddnum].count > boxarr[gndnum].count ?
	 VDD : GND);
   
   for (j = 0; j < 2; j++) {
      src_ln = (line == VDD ? vdd_line : gnd_line);
      bxnum = (line == VDD ? vddnum : gndnum);
      pending = NULL;
      lastgv = NULL;
      lastpx = -1;
      
      // sort spacers according to x position
      if (line == VDD)
         qsort(sprcells, sprcount, sizeof(cellptr), vddcellcmp);
      else
         qsort(sprcells, sprcount, sizeof(cellptr), gndcellcmp);

      /* Find ports that are connected to vdd and gnd.		*/
      /* Search for point-to-point connections.  Create 	*/
      /* channel lines if pt-to-pts cannot be made.		*/
      
      for (i = 0; i < total_ports; i++) {
	 cvals = portorder + i;
	 source = cvals->source;
	 tcell = *(cvals->source + cvals->pos);
	 pidx = cvals->port;
	 xpos = tcell->pos[pidx].x;
	 
	 if (tcell->net[pidx] != line || tcell->order == gnd_line ||
	       tcell->order == vdd_line)
	    continue;
	 
	 // try to route net with a point-to-point
	 if (gv_ptp(src_ln, bxnum, source, i, pidx))
	    continue;
	 
	 // lay a horizontal line in the channel
	 gvcell = find_closest_gv(src_ln, tcell, pidx);
	 if (gvcell == NULL) {
	    fprintf(stderr, "Failed to route order %d, port"
		  "\"%s\" to %s.\n", tcell->order,
		  boxarr[tcell->boxnum].ports[pidx].name,
		  (line == VDD ? "Vdd" : "GND"));
	    continue;
	 }
	 
	 if (lastpx > 0 && lastgv != NULL) {
	    newlength = abs(xpos - gvcell->pos[line].x);
	    tolastp = xpos - lastpx + PORT_SEP;
	    tolastgv = abs(xpos - lastgv->pos[line].x + PORT_SEP);
	    backlength = (tolastp < tolastgv ? tolastp : tolastgv);
	    if (lastgv->pos[line].x >= xpos + PORT_SEP)
	       backlength = tolastp;
	 }
	 
	 // attach port to last line if doing so makes the lines shorter
	 if (lastpx != -1 && lastgv != NULL && (backlength <= newlength ||
		  lastgv == gvcell)) {
	    if (pending->routed == 1)
	       endx = xpos;
	    else if (lastgv->pos[line].x > (lastpx - PORT_SEP) &&
		  lastgv->pos[line].x < (xpos + PORT_SEP)) {
	       // last gnd/vdd cell lies between the two ports
	       // turn last gv into gnd/vdd, unless gvcell has already been
	       // routed in a pt to pt and falls btw the ports)
	       if (!(gvcell->order == src_ln && gvcell->pos[line].x >
		     (lastpx - PORT_SEP) && gvcell->pos[line].x <
		     (xpos + PORT_SEP)) || gvcell == lastgv) {
		  convert_to_gv(lastgv, src_ln, bxnum);
		  endx = (lastgv->pos[line].x > xpos ? lastgv->pos[line].x :
			xpos);
	       }
	       else
		  endx = (gvcell->pos[line].x > xpos ? gvcell->pos[line].x :
			xpos);
	       // mark line as routed
	       pending->routed = 1;
	    }
	    
	    else if (gvcell->pos[line].x > (lastpx - PORT_SEP) && 
		  gvcell->pos[line].x < (xpos + PORT_SEP)) {
	       // turn current gvcell into gnd/vdd, mark line as routed
	       convert_to_gv(gvcell, src_ln, bxnum);
	       endx = (gvcell->pos[line].x > xpos ? gvcell->pos[line].x : xpos);
	    }
	    
	    else {
	       // search for spacer cell between two ports
	       lastgv = find_intermediate_cell(lastpx, xpos, src_ln);

	       if (lastgv != NULL) {
		  // route to intermediate cell (lastgv)
		  convert_to_gv(lastgv, src_ln, bxnum);
	          endx = (lastgv->pos[line].x > xpos ? lastgv->pos[line].x :
			xpos);
	       }
	       else
		  endx = xpos;
	    }
	    // extend line to include newest section
	    if (pending->end < endx + m1_width)
	       pending->end = endx + m1_width;
	 }
	 else {
	    // Route last line to gnd/vdd cell if neccessary.
	    if (netcount != first_gv && !(networks + netcount - 1)->routed)
	       convert_to_gv(((pending->begin - begin_gv->pos[line].x <
			lastgv->pos[line].x - pending->end) ? begin_gv :
			lastgv), src_ln, bxnum);
	    
	    // start new line
	    pending = create_gvnet(line, xpos);
	    begin_gv = gvcell;
	 }
	 
	 // set port's routed equal to the line number the port is connected to
	 // so if routed = 3, the port connects to *(v/glines + 3).
	 tcell->net[pidx] = netcount - 1;
	 lastgv = gvcell;
	 lastpx = xpos;
      }
      if (pending != NULL && !pending->routed)
         convert_to_gv(lastgv, src_ln, bxnum);

      if ((vdd_line && !gnd_line) || (gnd_line && !vdd_line))
	 break;
      line = ((line == GND) ? VDD : GND);  // move on to next net.
   }
   
   for (i = first_gv; i < netcount; i++)
      (networks + i)->routed = 0;  // so that lines will be placed in channels
   
   /* if leading spacers have been converted to gnd/vdd cells, layout's	*/
   /* offset has been changed and x-positions must be adjusted.		*/

   if ((final_offset = calc_initial_blanks()) != init_offset) {
      if (final_offset > init_offset)
	 fprintf(stderr, "\nProblems with leading gnd/vdd cells.\n");
      for (i = 2; i < netcount; i++) {
         (networks + i)->begin += init_offset - final_offset;
         (networks + i)->end += init_offset - final_offset;
      }
      for (i = 0; i < total_ports; i++) {
	 cvals = portorder + i;
	 (*(cvals->source + cvals->pos))->pos[cvals->port].x += init_offset -
	    final_offset;
      }
   }

}

void draw_gv_ptp(FILE *fm, cellptr cell1, cellptr cell2)
{
   int i, type, connection;
   cellptr gvcell = NULL, ocell;

   // check for ptp connections to gnd/vdd cells
   if (cell1->order == vdd_line || cell1->order == gnd_line) {
      if (cell2->order == vdd_line || cell2->order == gnd_line)
	 return;
      gvcell = cell1;
      ocell = cell2;
      connection = PTP_TO_LEFT;
   }
   else if (cell2->order == vdd_line || cell2->order == gnd_line) {
      gvcell = cell2;
      ocell = cell1;
      connection = PTP_TO_RIGHT;
   }
   if (gvcell == NULL)
      return;
   
   type = (gvcell->order == vdd_line ? VDD : GND);
   
   for (i = 0; i < boxarr[ocell->boxnum].numports; i++) {
      if (ocell->net[i] == type && ocell->routed[i] == connection) {
	 if (ocell == cell1)
	    outrecth(fm, ocell->pos[i].x, ocell->pos[i].y, gvcell->pos[type].x,
		  m1_width);
	 else
	    outrecth(fm, gvcell->pos[type].x, ocell->pos[i].y,
		  ocell->pos[i].x, m1_width);
      }
   }
}
