/*----------------------------------------------------------------------*/
/* graphics.c								*/
/*									*/
/* Draw routes on simple X window					*/
/*									*/
/* Copyright 2001 - Tim Edwards						*/
/*----------------------------------------------------------------------*/

#include "defines.h"

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

Display *dpy = NULL;
Window win;
Colormap cmap;
GC gc;
Dimension width, height;
XFontStruct *font_info;

extern int m1_width;

extern void route_current();

/*--------------------------------------------------------------*/      
/* Find ll x coordinate of all cells to print route layout.     */      
/* Return lengths of upper and lower rows.                      */      
/*--------------------------------------------------------------*/      

void calc_cell_offsets(int **loff, int **uoff, int *llength, int *ulength)
{
   int i, lx = 0, ux = 0, *lptr, *uptr;

   lptr = *loff = (int *)malloc(sizeof(int) * cols);
   uptr = *uoff = (int *)malloc(sizeof(int) * cols);

   for (i = 0; i < cols; i++) {
      lptr[i] = lx;
      uptr[i] = ux;
      lx += boxarr[(*(lcells + i))->boxnum].w;
      ux += boxarr[(*(ucells + i))->boxnum].w;
   }
   *llength = lx;
   *ulength = ux;
}

/*--------------------------------------------------------------*/
/* Initialize drawing by opening the X window 			*/
/*--------------------------------------------------------------*/

#define WWIDTH 800
#define WHEIGHT 500

int open_window()
{
   char *fontname = "9x15";
   unsigned long blackpix;
   unsigned long whitepix;
   XEvent event;
   XConfigureEvent *ConfigureEvent;
   int *uoffsets, *loffsets, llength, ulength;

   calc_cell_offsets(&loffsets, &uoffsets, &llength, &ulength);
   width = MAX(llength, ulength) + 100;
   height = WHEIGHT;

   dpy = XOpenDisplay(NULL);

   blackpix = BlackPixel(dpy, DefaultScreen(dpy));
   whitepix = WhitePixel(dpy, DefaultScreen(dpy));

   win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 20, 20,
		width, height, 1, blackpix, whitepix);
   gc = XCreateGC(dpy, win, 0, NULL);

   /* Set GC values one at a time */

   if ((font_info = XLoadQueryFont(dpy,fontname)) == NULL) {
      fprintf(stderr, "Cannot open 9x15 font\n");
      quit(1);
   }  

   XSetFont(dpy, gc, font_info->fid);
   XSetForeground(dpy, gc, blackpix);
   XSetBackground(dpy, gc, whitepix);

   XSetLineAttributes (dpy, gc, 1, LineSolid, CapRound, JoinRound);
   XSetFillStyle (dpy, gc, FillSolid);

   XClearArea(dpy, win, 0, 0, width, height, 0);
   XMapWindow(dpy, win);

   return (int)width;
}

/*--------------------------------------------------------------*/
/* Close the X-Window						*/
/*--------------------------------------------------------------*/

void close_window()
{
   if (dpy != NULL) XCloseDisplay(dpy);
}

/*--------------------------------------------------------------*/
/* Draw a line							*/
/*--------------------------------------------------------------*/

void UDrawLine(FILE *fo, int x1, int y1, int x2, int y2)
{
   if (fo == stderr) {	/* Draw to X window */
      XDrawLine(dpy, win, gc, x1, y1, x2, y2);
   }
   else {		/* Write PostScript to file */
      fprintf(fo, "%d %d moveto %d %d lineto stroke\n",
		x1, 700 - y1, x2, 700 - y2);
   }
}

/*--------------------------------------------------------------*/
/* Write text							*/
/*--------------------------------------------------------------*/

void UDrawText(FILE *fo, int x1, int y1, char *textstr, int len)
{
   if (fo == stderr) {	/* Draw to X window */
      XDrawString(dpy, win, gc, x1, y1, textstr, len);
   }
   else {		/* Write PostScript to file */
      fprintf(fo, "%d %d moveto (%s) show\n", x1, 700 - y1, textstr);
   }
}

/*--------------------------------------------------------------*/
/* Compute text length (if necessary) and write text		*/
/*--------------------------------------------------------------*/

void UDrawTextS(FILE *fo, int x1, int y1, char *textstr)
{
   int len = 0;

   if (fo == stderr) len = strlen(textstr);
   UDrawText(fo, x1, y1, textstr, len);
}

/*--------------------------------------------------------------*/
/* Check for cross-channel connection				*/
/*--------------------------------------------------------------*/

int check_across(int thisnet)
{
   cellptr thiscell;
   int i, j, numports;

   /* If the net is in the port list for at least one upper	*/
   /* row cell and at least one lower row cell, then the route	*/
   /* must go across the channel.				*/

   for (i = 0; i < cols; i++) {
      thiscell = *(ucells + i);
      if (thiscell->order == 0) continue;
      numports = boxarr[thiscell->boxnum].numports;
      for (j = 0; j < numports; j++)
         if (thiscell->net[j] == thisnet)
	    break;
      if (j < numports) break;
   }
   if (i == cols) return 0;

   for (i = 0; i < cols; i++) {
      thiscell = *(lcells + i);
      if (thiscell->order == 0) continue;		/* spacer cell */
      numports = boxarr[thiscell->boxnum].numports;
      for (j = 0; j < numports; j++)
         if (thiscell->net[j] == thisnet) return 1;
   }
   return 0;
}

/*--------------------------------------------------------------*/
/* Draw connections directly across the cell			*/
/*--------------------------------------------------------------*/

void draw_across(FILE *fo, int yoff1, int yoff2, int xoffset)
{
   int i, j, numports;
   netptr thisnet;
   cellptr thiscell;

   for (i = 0; i < cols; i++) {
      thiscell = *(ucells + i);
      if (thiscell->order == 0) continue;
      numports = boxarr[thiscell->boxnum].numports;
      for (j = 0; j < numports; j++) {
	 thisnet = networks + thiscell->net[j];
	 if (thisnet->channel == -1)
	    if (check_across(thiscell->net[j]))
	       UDrawLine(fo, thiscell->pos[j].x,
			yoff1, thiscell->pos[j].x, yoff2);
      }
   }
}

/*--------------------------------------------------------------*/
/* Draw internal cell-to-cell connections			*/
/*--------------------------------------------------------------*/

void draw_internal(FILE *fo, int yoff1, int yoff2, int xoffset, cellptr ocell)
{
   int i, o, numports, begx;
   netptr thisnet;

   if (ocell->order == 0) return;
   numports = boxarr[ocell->boxnum].numports;

   for (i = 0; i < numports; i++) {
      o = (ocell->orient ? numports - i - 1 : i);
      if (ocell->net[o] >= 2) {
	 thisnet = networks + ocell->net[o];
	 if (thisnet->channel == -1) {
	    if (!check_across(ocell->net[o])) {
	       /* This shouldn't be happening---check for an	*/
	       /* error elsewhere?				*/
	       begx = MIN(ocell->pos[o].x, thisnet->begin);
	       UDrawLine(fo, begx, yoff1, begx, yoff2);
	       UDrawLine(fo, thisnet->end - m1_width,
			yoff1, thisnet->end - m1_width, yoff2);
	       UDrawLine(fo, begx,
			yoff1, thisnet->end - m1_width, yoff1);
	    }
	 }
	 else if (ocell->pos[o].x > thisnet->end) {
	    UDrawLine(fo, thisnet->end - m1_width,
		yoff1, thisnet->end - m1_width, yoff2);
	    UDrawLine(fo, ocell->pos[o].x,
		yoff1, ocell->pos[o].x, yoff2);
	    UDrawLine(fo, ocell->pos[o].x,
		yoff1, thisnet->end - m1_width, yoff1);
	 }
	 else if (ocell->pos[o].x < thisnet->begin) {
	    UDrawLine(fo, thisnet->begin,
		yoff1, thisnet->begin, yoff2);
	    UDrawLine(fo, ocell->pos[o].x,
		yoff1, ocell->pos[o].x, yoff2);
	    UDrawLine(fo, thisnet->begin,
		yoff1, ocell->pos[o].x, yoff1);
	 }
      }
   }
}

/*--------------------------------------------------------------*/
/* Draw connections to power planes (Vdd or GND)		*/
/*--------------------------------------------------------------*/

void draw_power(FILE *fo, int yoff1, int yoff2, int xoffset, cellptr ocell, int nettype)
{
   int i, o, numports;

   if (ocell->order == 0)
      return;
   numports = boxarr[ocell->boxnum].numports;

   for (i = 0; i < numports; i++) {
      o = (ocell->orient ? numports - i - 1 : i);
      if (ocell->net[o] == nettype) {
	 UDrawLine(fo, xoffset + ocell->pos[o].x,
		yoff1, xoffset + ocell->pos[o].x, yoff2);
      }
   }
}

/*--------------------------------------------------------------*/
/* Draw port and net names for each cell port			*/
/*--------------------------------------------------------------*/

void draw_ports(FILE *fo, int yoff1, int yoff2, int xoffset, cellptr ocell)
{
   int i, o, numports;
   static char chrout[2] = "x";

   numports = boxarr[ocell->boxnum].numports;
   if (ocell->order == 0)
      return;
   else if (ocell->order == gnd_line || ocell->order == vdd_line) {
      o = (ocell->order == vdd_line);
      *chrout = (o) ? 'V' : 'G';
      UDrawText(fo, xoffset + ocell->pos[o].x, yoff1, chrout, 1);

      *chrout = (o) ?  'b' : 'a';
      UDrawText(fo, xoffset + (*(ucells + i))->pos[o].x, yoff2, chrout, 1);
   }
   else {
      for (i = 0; i < numports; i++) {
         o = (ocell->orient ? numports - i - 1 : i);
	 *chrout = boxarr[ocell->boxnum].ports[o].name[0];
	 UDrawText(fo, xoffset + ocell->pos[o].x, yoff1, chrout, 1);

	 *chrout = netchar(ocell->net[o]);
	 UDrawText(fo, xoffset + ocell->pos[o].x, yoff2, chrout, 1);
      }
   }
}

/*--------------------------------------------------------------*/
/* Draw the stick-diagram layout in an X-window			*/
/*--------------------------------------------------------------*/

void draw_layout(chstructptr channels, int num_channels, FILE *fileout)
{
   int i, j, k, m, o, cpos, lpos, numports;
   int *uoffsets, *loffsets, llength, ulength;
   int maxl, net, offset; // offset=measure of blank cells at beginning
   int src_ln, yoffset, ytop, ybottom, thisnetnum;
   chstructptr thischannel;
   portinfoptr cvals;
   cellptr tcell;
   netptr thisnet;

   if (fileout == NULL) return;		/* no drawing */
   
   calc_cell_offsets(&loffsets, &uoffsets, &llength, &ulength);
   offset = calc_initial_blanks();
   
   ytop = 65;
   ybottom = ytop + 10 * num_channels + 10;
   yoffset = ytop;

   XClearArea(dpy, win, 0, 0, width, height, 0);

   for (i = 0; i < cols; i++) {
      UDrawLine(fileout, offset, ytop - 50, ulength, ytop - 50);
      draw_power(fileout, ytop - 50, ytop - 30, offset, *(ucells + i), VDD);
      draw_internal(fileout, ytop - 40, ytop - 30, offset, *(ucells + i)); 

      draw_ports(fileout, ytop, ytop - 16, offset, *(ucells + i));
      draw_ports(fileout, ybottom + 16, ybottom + 32, offset, *(lcells + i));

      draw_internal(fileout, ybottom + 50, ybottom + 40, offset, *(lcells + i)); 
      draw_power(fileout, ybottom + 60, ybottom + 40, offset, *(lcells + i), GND);
      UDrawLine(fileout, offset, ybottom + 60, llength, ybottom + 60);
   }
   draw_across(fileout, ytop, ybottom, offset);

   for (k = 0; k < num_channels; k++) {
      thischannel = channels + k;
      yoffset += 10;

      for (i = 0; i < thischannel->num_nets; i++) {
	 thisnetnum = *(thischannel->nets + i);
	 thisnet = networks + thisnetnum;
	 UDrawLine(fileout, thisnet->begin, yoffset,
		thisnet->end - m1_width, yoffset);

	 for (j = 0; j < cols; j++) {
	    tcell = *(ucells + j);
	    if (tcell->order == 0) continue;
	    numports = boxarr[tcell->boxnum].numports;
	    for (m = 0; m < numports; m++) {
	       if (tcell->net[m] == thisnetnum &&
			thisnet->begin <= tcell->pos[m].x &&
			thisnet->end >= tcell->pos[m].x) {
		  UDrawLine(fileout, tcell->pos[m].x, yoffset,
			tcell->pos[m].x, ytop);
	       }
	    }
	    tcell = *(lcells + j);
	    if (tcell->order == 0) continue;
	    numports = boxarr[tcell->boxnum].numports;
	    for (m = 0; m < numports; m++) {
	       if (tcell->net[m] == thisnetnum &&
			thisnet->begin <= tcell->pos[m].x &&
			thisnet->end >= tcell->pos[m].x) {
		  UDrawLine(fileout, tcell->pos[m].x, yoffset,
			tcell->pos[m].x, ybottom);
	       }
	    }
	 }
      }
   }

   free(uoffsets);
   free(loffsets);
   XSync(dpy, 0);
}

/*--------------------------------------------------------------*/
