/*----------------------------------------------------------------------*/
/* magic.c --- Generate .mag layout for the supplied route		*/
/*									*/
/* Copyright 2001 - Jennifer Roberts and Tim Edwards			*/
/*----------------------------------------------------------------------*/

#include "defines.h"

/*------------------------------*/
/* Externally-defined routines	*/
/*------------------------------*/

extern boxptr read_box(char *);

/*------------------------------*/
/* Global external variables    */
/*------------------------------*/

extern int netcount;
extern netptr networks;
extern int cols;
extern cellptr *lcells, *ucells;
extern char techname[30];
extern boxptr boxarr;
extern int boxcount;
extern int fet_design;

extern int total_ports;
extern portinfoptr portorder;

extern int m1_width;               // width of a m1 line
extern int m2_width;               // width of a m2 line
extern int via_width;              // width of a via
extern int via_height;             // height of a via
extern int m1_space;               // spacing between m1 lines
extern int m2_space;               // spacing between m2 lines
extern int contact_size;           // contact size for a substrate/well contact

/*--------------------------------------------------------------*/
/* Generate a rect (rectangle) record in the .mag file		*/
/*--------------------------------------------------------------*/

void outrect(FILE *fm, int x0, int y0, int width, int height)
{
   fprintf(fm, "rect %d %d %d %d\n", x0, y0, x0 + width, y0 + height);
}

/*--------------------------------------------------------------*/
/* Generate a rect (rectangle) record in the .mag file for	*/
/* a horizontally-oriented wire of width "width"		*/
/*--------------------------------------------------------------*/

void outrecth(FILE *fm, int x0, int y0, int x1, int width)
{
   if (x0 == x1) return;
   fprintf(fm, "rect %d %d %d %d\n", x0, y0, x1, y0 + width);
}

/*--------------------------------------------------------------*/
/* Generate a rect (rectangle) record in the .mag file for      */
/* a vertically-oriented wire of width "width"                  */
/*--------------------------------------------------------------*/

void outrectv(FILE *fm, int x0, int y0, int width, int y1)
{
   if (y0 == y1) return;
   fprintf(fm, "rect %d %d %d %d\n", x0, y0, x0 + width, y1);
}

/*--------------------------------------------------------------*/
/* Draw a box of type "tbox" at (ll) x0, y0; add gnd and vdd	*/
/* connections.  Log y positions of ports.			*/
/*--------------------------------------------------------------*/

void draw_box(FILE *fm, cellptr thiscell, int x0, int y0, boxptr tbox)
{
   int i, orient, xoff, yoff, overlap;

   if (tbox->w == 0)	// box has no width; ignore it
      return;

   yoff = y0 - tbox->y;	// subtract tbox->ht if flipped in y; not implemented

   if (thiscell) {
      overlap = thiscell->overlap;
      if (thiscell->orient) {
	 orient = -1;
	 xoff = x0 + tbox->w + tbox->x;
      }
      else {
	 orient = 1;
	 xoff = x0 - tbox->x;
      }
   }
   else {
      overlap = 0;
      orient = 1;
      xoff = x0 - tbox->x;
   }

   fprintf(fm, "use %s %s_%d\n", tbox->name, tbox->name, tbox->instance);
   (tbox->instance)++;
   fprintf(fm, "timestamp 0\n");
   fprintf(fm, "transform %d 0 %d 0 1 %d\n", orient, xoff, yoff);
   fprintf(fm, "box 0 0 %d %d\n", tbox->w, tbox->ht);
   
   if (thiscell == NULL || thiscell->order == 0)
      return;  // cap cell or spacer cell

   if (thiscell->order == vdd_line || thiscell->order == gnd_line) {
      thiscell->pos[GND].y = (gndnum ? y0 + boxarr[gndnum].gnd.ll.y : 0);
      thiscell->pos[VDD].y = (vddnum ? y0 + boxarr[vddnum].vdd.ll.y : 0);
      return;
   }
   
   for (i = 0; i < tbox->numports; i++) {
      // initialize y position of port
      thiscell->pos[i].y = y0 + tbox->ports[i].ll.y;
      
      // route connections in fet_designs
      if (fet_design && thiscell->net[i] == GND) {
	 fprintf(fm, "<< metal1 >>\n");
	 outrectv(fm, thiscell->pos[i].x, y0 + tbox->gnd.ur.y, m1_width, 
	       thiscell->pos[i].y);

	 // add GND/Vdd connection to fet_designs
	 if ((tbox->ports[i].idx == ISOURCE) && !overlap) {
	    outrectv(fm, thiscell->pos[i].x, thiscell->pos[i].y,
		  m1_width, tbox->ht - 4);
	    fprintf(fm, "<< psubstratepcontact >>\n");
	    outrect(fm, thiscell->pos[i].x - (contact_size - m1_width) / 2,
		  tbox->ht - contact_size, contact_size, contact_size);
	 }
      }
	 
      // route Vdd connections in fet_designs
      else if (fet_design && thiscell->net[i] == VDD) {
	 fprintf(fm, "<< metal1 >>\n");
	 outrectv(fm, thiscell->pos[i].x, thiscell->pos[i].y + m1_width,
	    m1_width, y0 + tbox->vdd.ll.y);
	 // add GND/Vdd connection to fet_designs
	 if ((tbox->ports[i].idx == ISOURCE) && !overlap) {
	    outrectv(fm, thiscell->pos[i].x, y0 + 4, m1_width,
		  thiscell->pos[i].y + m1_width);
	    fprintf(fm, "<< nsubstratencontact >>\n");
	    outrect(fm, thiscell->pos[i].x - (contact_size - m1_width) / 2,
		y0, contact_size, contact_size);
	 }
      }
   }
}

/*--------------------------------------------------------------*/
/* Point-to-point route:					*/
/* Connects port0 of "thiscell" to port1 of "nextcell"		*/
/*--------------------------------------------------------------*/

void point_to_point(FILE *fm, cellptr thiscell, cellptr nextcell,
		int port0, int port1)
{
   int net = thiscell->net[port0];
   
   if (thiscell->pos[port0].x < nextcell->pos[port1].x) // ignore overlap sources
      outrecth(fm, thiscell->pos[port0].x, thiscell->pos[port0].y,
	    nextcell->pos[port1].x + m1_width, m1_width);
}

/*--------------------------------------------------------------*/
/* Check to see whether point to point connections can be made	*/
/* between adjacent cells.					*/
/*--------------------------------------------------------------*/

int check_for_ptp(cellptr cell1, cellptr cell2, char direc1, char direc2,
      int idx1, int *idx2, int row)
{
   boxptr bx1 = boxarr + cell1->boxnum, bx2 = boxarr + cell2->boxnum;
   int i, ypos;
   int route1, route2;	// tell whether cell can be routed in correct direction
   			// for point to point route
   
   if (cell1->order == 0 || cell2->order == 0 ||
	 cell1->order == gnd_line || cell2->order == gnd_line ||
	 cell1->order == vdd_line || cell2->order == vdd_line)
      return 0;

   route1 = (direc1 == 'e' ? bx1->ports[idx1].east : bx1->ports[idx1].west);
   if (route1 == 0 || (route1 == IF_ON_TOP && row == LOWER) ||
	 (route1 == IF_ON_BOTTOM && row == UPPER))
      return 0;
   
   ypos = boxarr[cell1->boxnum].ports[idx1].ll.y;
   for (i = 0; i < bx2->numports; i++) {
      if (boxarr[cell2->boxnum].ports[i].ll.y == ypos) {
	 route2 = (direc2 == 'e' ? bx2->ports[i].east : bx2->ports[i].west);
	 if (route2 != 0 && !(route2 == IF_ON_TOP && row == LOWER) &&
	    !(route2 == IF_ON_BOTTOM && row == UPPER) && cell1->net[idx1] ==
	    cell2->net[i]) {
	    *idx2 = i;
	    return 1;
	 }
      }
   }
   
   return 0;
}

/*--------------------------------------------------------------*/
/* Searches for a port in the row pointed to by *osrc that is	*/
/* located at x0 and attached to "net".	 Returns a point 	*/
/* containing location of the matching port if such a port	*/
/* exists, NULL otherwise.					*/
/*--------------------------------------------------------------*/

point *matching_v(int x0, int net, cellptr *osrc, int startpos)
{
   int i, j, numports, l, r;
   int direc = 1;  // direction to search through cells: 1 = right, -1 = left
   cellptr ocell;

   for (j = startpos; j >= 0 && j < cols; ) {
      ocell = *(osrc + j);
      if (ocell->order == 0 || ocell->order == gnd_line ||
	    ocell->order == vdd_line) {

	 // skip blanks and check for vertical gnd/vdd connections

	 if (vdd_line && ocell->order == vdd_line && net == ocell->net[VDD]
	       && abs(ocell->pos[VDD].x - x0) < PORT_SEP) {
	       return &(ocell->pos[VDD]);
	 }

	 if (gnd_line && ocell->order == gnd_line && net == ocell->net[GND]
	       && abs(ocell->pos[GND].x - x0) < PORT_SEP) {
	       return &(ocell->pos[GND]);
	 }
	 if (j == cols - 1)
	    direc = -1;
	 j += direc;
	 continue;
      }
      numports = boxarr[ocell->boxnum].numports;
      l = (ocell->orient ? numports - 1 : 0);
      r = (ocell->orient ? 0 : numports - 1);
      if (ocell->pos[l].x <= (x0 - PORT_SEP) && ocell->pos[r].x <=
	    (x0 - PORT_SEP)) {
	 // all of ocell's ports are to the left of x0
	 if (direc == -1)
	    break; // x0 lies between the ports of 2 adjacent cells
	 j++;  // continue searching to the right
      }
      else if (ocell->pos[l].x >= (x0 + PORT_SEP) && ocell->pos[r].x
	    >= (x0 + PORT_SEP)) {
	 // all of ocell's ports are to the right of x0
	 direc = -1;
	 j--;  // continue searching to the left
      }
      else {  // x0 lies within ocell -- search ocell for port located at x0
	 for (i = 0; i < numports; i++) {
	    if (abs(ocell->pos[i].x - x0) < PORT_SEP && ocell->net[i] == net) {
	       return &(ocell->pos[i]);	// port at x0 has matching net found
	    }
	 }
	 if ((x0 >= ocell->pos[0].x && x0 <= ocell->pos[numports - 1].x) ||
	       (x0 <= ocell->pos[0].x && x0 >= ocell->pos[numports - 1].x))
	    break;
	 j += direc;
      }
   }
   return NULL;
}

/*--------------------------------------------------------------*/
/* Generate the detailed layout for vertical connections to	*/
/* the networks, by filling in a vsegment record for each	*/
/* vertical route line (not counting GND/Vdd connections).	*/
/* Return 0 on success, 1 if layout could not be properly	*/
/* routed.							*/
/*--------------------------------------------------------------*/

int vroute_prep(portinfo *cvals, chstructptr channels,
	int num_channels, vsegmentptr vseg)
{
   int icol = cvals->pos;
   int irow = ((cvals->source == ucells) ? UPPER : LOWER);
   int iport = cvals->port;

   cellptr thiscell;
   int net, py, cy, x0, y0, ychan;
   point *matchpos;
   chstructptr thischan, midchan;
   
   thiscell = *(cvals->source + icol);
   x0 = thiscell->pos[iport].x;
   py = thiscell->pos[iport].y;
   net = thiscell->net[iport];
   ychan = (networks + net)->channel;
   thischan = channels + ychan;
   midchan = channels + (num_channels >> 1);
   
   /* If the net is Vdd or GND, ignore. */
   /* If the net is listed as already routed, ignore */

   if (net < 2) return -1;
   else if (thiscell->routed[iport] == 1) return -1;

   vseg->net = net;
   vseg->metal = (fet_design) ? METAL1 : METAL2;
   vseg->yjog = -1;	  /* initial value (no jog)			*/
   vseg->contact.x = -1;  /* initial value (no contact to channel)	*/
   vseg->contact.y = -1;  /* initial value (no contact to channel)	*/

   // If port's net is not assigned to a channel or if port is not located above
   // or below the channel it's assigned to, see if a strictly vertical
   // connection is needed.  only continue if a vertical route is neccessary.

   if ((ychan < 0) || (x0 < (networks + net)->begin) ||
		(x0 > (networks + net)->end))  {

      /* make sure there is a port directly above/below iport that 	*/
      /* shares the same net.  Otherwise, this connection does not need	*/
      /* a vertical route (return -1 to the caller to signal this fact)	*/

      if ((matchpos = matching_v(x0, net, (irow ? lcells : ucells), icol)) == NULL)
	 return -1;

      /* strictly vertical route.  Do for irow = LOWER only, so the route */
      /* doesn't get duplicated.					 */

      if (irow == UPPER) return -1;

      vseg->l.x = x0;
      vseg->l.y = py;
      vseg->u.x = matchpos->x;
      vseg->u.y = matchpos->y;
      if (x0 != matchpos->x)
	 vseg->yjog = midchan->yval;  /* place jog in middle */
   }
   else {

      /* cy is the point where the route connects to the channel */
      cy = thischan->yval;
      vseg->l.x = vseg->u.x = x0;
      vseg->l.y = MIN(py, cy);
      vseg->u.y = MAX(py, cy);
      vseg->contact.x = x0;	/* adjust to center the contact? */
      vseg->contact.y = cy;
      vseg->yjog = cy;
   }
   return 0;
}

/*--------------------------------------------------------------*/
/* Resolve tight spots in the vertical routing by adding dogleg	*/
/* routes to the vertical route lines.  Adjust positions 	*/
/* forward and backward as necessary to maintain design rule	*/
/* spacing, and crop or extend horizontal routes to match.	*/
/*								*/
/* Return 0 if doglegs have been resolved, -1 if no resolution	*/
/* is possible.							*/
/*--------------------------------------------------------------*/

int dogleg(vsegmentptr vsegs, int num_vsegs, chstructptr channels,
	int num_channels)
{
   int i, expand, lastx, nextx, ycase, slack, uusep, llsep, ulsep, lusep;
   int uucond, llcond, ulcond, lucond;

   for (i = 0; i < num_vsegs - 1; i++) {

      /* Always check against next route to the right */

      uusep = (vsegs[i + 1].u.x - vsegs[i].u.x) - PORT_SEP;
      llsep = (vsegs[i + 1].l.x - vsegs[i].l.x) - PORT_SEP;
      ulsep = (vsegs[i + 1].u.x - vsegs[i].l.x) - PORT_SEP;
      lusep = (vsegs[i + 1].l.x - vsegs[i].u.x) - PORT_SEP;

      uucond = (uusep < 0);
      llcond = (llsep < 0);
      ulcond = (ulsep < 0) && (vsegs[i+1].yjog - vsegs[i].yjog - PORT_SEP) < 0;
      lucond = (lusep < 0) && (vsegs[i].yjog - vsegs[i+1].yjog - PORT_SEP) < 0;

      /* If upper segments and lower segments are far enough apart, but */
      /* the doglegs overlap, then separate the doglegs.		*/

      if ((ulcond || lucond) && !(uucond) && !(llcond)) {
	 /* Are these conditions right?? */
	 if (ulcond) {
	    if (vsegs[i].yjog > vsegs[i].l.y)
	       vsegs[i].yjog = vsegs[i+1].yjog - PORT_SEP;
	    else
	       vsegs[i+1].yjog = vsegs[i].yjog + PORT_SEP;
	 }
	 else {	 /* lucond */
	    if (vsegs[i+1].yjog > vsegs[i+1].l.y)
	       vsegs[i+1].yjog = vsegs[i].yjog - PORT_SEP;
	    else
	       vsegs[i].yjog = vsegs[i+1].yjog + PORT_SEP;
	 }
      }
      else if (uucond || llcond || ulcond || lucond) {

	 /* There is only a conflict if the Y values are also too close */
	 /* or overlap							*/

	 if ((vsegs[i + 1].l.y - CHAN_SEP < vsegs[i].u.y) &&
	     	(vsegs[i + 1].u.y + CHAN_SEP > vsegs[i].l.y)) {

	    /* requires expanding the gap between the nets by this distance */
	    expand = PORT_SEP - MIN((vsegs[i + 1].l.x - vsegs[i].u.x),
			(vsegs[i + 1].u.x - vsegs[i].l.x));

	    /* If both Y values attach to the same channel, then there	*/
	    /* is no conflict, but we should dogleg one of them so that	*/
	    /* the vias align.						*/

	    if (vsegs[i+1].net == vsegs[i].net) {
	       if (vsegs[i].l.y == vsegs[i+1].u.y) {
	          vsegs[i].l.x = vsegs[i+1].u.x;
		  vsegs[i].yjog = vsegs[i+1].u.y;
	       }
	       else {
	          vsegs[i].u.x = vsegs[i+1].l.x;
		  vsegs[i].yjog = vsegs[i+1].l.y;
	       }
	    }

	    /* For now, only treat those conflicts with badness > m2_width	*/
	    /* Would like to fix this in the future.				*/

	    else if (expand > /* m2_width */ 0) {
	       fprintf(stderr, "Route conflict at x = %d\n", vsegs[i].u.x);
	       fprintf(stderr, "  Needs to be expanded by %d units\n", expand);
	       fprintf(stderr, "  Vertical route = %d\n", i + 1);
	       fprintf(stderr, "  X1 = %d, Y1min = %d, Y1max = %d\n",
		   vsegs[i].u.x, vsegs[i].l.y, vsegs[i].u.y);
	       fprintf(stderr, "  X2 = %d, Y2min = %d, Y2max = %d\n",
		   vsegs[i+1].u.x, vsegs[i+1].l.y, vsegs[i+1].u.y);

	       /* To be done:  Let this routine plow backwards and forwards	*/
	       /* as many places as necessary to satisfy DRC, or until routing	*/
	       /* appears to be impossible.					*/

	       /* For now, look one place backwards, then one place forwards	*/

	       /* case 0:   |		case 1:   |			*/
	       /*	   ||			  ||			*/
	       /*	   | 			   |			*/

	       ycase = (vsegs[i+1].u.y > vsegs[i].u.y) ? 0 : 1;
	       if (i > 0) {
	          lastx = MAX(vsegs[i-1].l.x, vsegs[i-1].u.x);
		  slack = MIN(vsegs[i].l.x, vsegs[i].u.x) - lastx - PORT_SEP;
		  if (slack > expand) slack = expand; 
		  /* For now, add the condition that we can only jog routes */
		  /* which are not already jogged.			    */
		  if ((slack > 0) && (vsegs[i].u.x == vsegs[i].l.x)) {   
		     if (ycase == 0) {
		        vsegs[i].u.x -= slack;
			vsegs[i].yjog = vsegs[i+1].l.y - CHAN_SEP;
		     }
		     else {
			vsegs[i].l.x -= slack;
			vsegs[i].yjog = vsegs[i+1].u.y + CHAN_SEP;
		     }
		     expand -= slack;
		  }
	       }
	       if ((expand > 0) && (i < num_vsegs - 2)) {
	          nextx = MIN(vsegs[i+2].l.x, vsegs[i+2].u.x);
		  slack = nextx - MAX(vsegs[i+1].l.x, vsegs[i+1].u.x) - PORT_SEP;
		  if (slack > expand) slack = expand; 
		  /* For now, add the condition that we can only jog routes */
		  /* which are not already jogged.			    */
		  if ((slack > 0) && (vsegs[i+1].u.x == vsegs[i+1].l.x)) {   
		     if (ycase == 0) {
		        vsegs[i+1].l.x += slack;
		        vsegs[i+1].yjog = vsegs[i].u.y + CHAN_SEP;
		     }
		     else {
		        vsegs[i+1].u.x += slack;
		        vsegs[i+1].yjog = vsegs[i].l.y - CHAN_SEP;
		     }
		     expand -= slack;
		  }
	       }
	       if (expand > 0) {
		  fprintf(stderr, "Tried to dogleg route but ran out of space!\n");
	       }
	    }
	 }
      }
   }
}

/*--------------------------------------------------------------*/
/* Write a vertical route layout to the magic file for standard */
/* cells (fet_design == 0).  Route metal2 vertically, metal1	*/
/* horizontally.  This is much simpler than the FET design	*/
/* routing, as we do not try to simplify connectivity by	*/
/* routing metal1 vertically to avoid unnecessary placing of	*/
/* vias in the signal path.					*/
/*--------------------------------------------------------------*/

int vroute2(FILE *fm, vsegmentptr thisseg, chstructptr channels, int num_channels)
{
   int vx, vy;

   /* Route all metal 2 */

   fprintf(fm, "<< metal2 >>\n");

   if (thisseg->yjog == -1) {		/* no jog */
      outrectv(fm, thisseg->l.x, thisseg->l.y, m2_width, thisseg->u.y);
   }
   else if (thisseg->yjog != -1) {	/* jogged */
      outrectv(fm, thisseg->l.x, thisseg->l.y, m2_width, thisseg->yjog + m2_width);
      outrectv(fm, thisseg->u.x, thisseg->yjog, m2_width, thisseg->u.y);
      if (thisseg->u.x > thisseg->l.x + m2_width)
         outrecth(fm, thisseg->l.x + m2_width, thisseg->yjog, thisseg->u.x, m2_width);
      else if (thisseg->u.x < thisseg->l.x - m2_width)
         outrecth(fm, thisseg->u.x + m2_width, thisseg->yjog, thisseg->l.x, m2_width);

      /* Check if the contact position needs to be moved to correspond to the jog */

      /* Case 1: Contact is below the jog */
      if ((thisseg->contact.x == thisseg->l.x)
			&& (thisseg->yjog < thisseg->contact.y))
	 thisseg->contact.x = thisseg->u.x;

      /* Case 2: Contact is above the jog */
      else if ((thisseg->contact.x == thisseg->u.x)
			&& (thisseg->yjog > thisseg->contact.y))
	 thisseg->contact.x = thisseg->l.x;
   }

   /* Position I/O contact on the channel */

   vx = VIA_X(thisseg->contact.x);
   vy = VIA_Y(thisseg->contact.y);
   fprintf(fm, "<< via >>\n");
   outrect(fm, vx, vy, via_width, via_height);

   /* Add Input/Output labels to via, if the network is named in the	*/
   /* schematic ("int" and "ext" names are not considered (xcircuit	*/
   /* convention).							*/

   if (((networks + thisseg->net)->name != NULL) &&
         (strstr((networks + thisseg->net)->name, "int") == NULL) &&
         (strstr((networks + thisseg->net)->name, "ext") == NULL)) {

      fprintf(fm, "<< labels >>\n");
      fprintf(fm, "rlabel via %d %d %d %d 7 %s\n", vx, vy, vx + via_width,
            vy + via_height, (networks + thisseg->net)->name);
      /* fprintf(fm, "port %d e\n", ...); */ /* to be completed. . . */

      /* De-allocate names as we go, so that we don't repeat an I/O label */

      free((networks + thisseg->net)->name);
      (networks + thisseg->net)->name = NULL;
   }
}

/*--------------------------------------------------------------*/
/* Write a vertical route layout to the magic file.  Generate	*/
/* "jumpers" in metal2 across horizontal channel routes where	*/
/* necessary.  Place contacts where necessary.			*/
/*--------------------------------------------------------------*/
   
int vroute(FILE *fm, vsegmentptr thisseg, chstructptr channels, int num_channels)
{
   int i, k, jl, jh, curx, cury, lx, ly, ux, uy, y0, x0, vx, vy;
   chstructptr thischan;

   /* Check for need to jumper with metal2 over unconnected channels */

   jl = num_channels;
   jh = -1;

   for (i = 0; i < num_channels; i++) {
      thischan = channels + i;
      y0 = thischan->yval;
      x0 = (y0 > thisseg->yjog) ? thisseg->u.x : thisseg->l.x;
      if (y0 < thisseg->u.y && y0 > thisseg->l.y && y0 != thisseg->contact.y) {

	 /* look for nets which cross or come too close to x0 value */

         for (k = 0; k < thischan->num_nets; k++) {
            int knet = *(thischan->nets + k);
            if ((knet != thisseg->net) &&
			((networks + knet)->begin < (x0 + m1_width + m1_space)) &&
                	((networks + knet)->end > (x0 - m1_space))) {
               if (i < jl) jl = i;
	       if (i > jh) jh = i;
	       break;
	    }
         }
      }
   }
   /* move jl and jh out by one channel spacing */
   if (jh != -1) jh++;
   if (jl != num_channels) jl--;

   /* Set y positions for each jumper end contact */

   if (jl == -1) {			/* top (above all channels) */
      thischan = channels;
      uy = thischan->yval + m1_width + BOX_TO_CHAN;
      if (uy > thisseg->u.y - via_height)
	 uy = thisseg->u.y - via_height;
   }
   else if (jl != num_channels) {
      thischan = channels + jl;
      uy = thischan->yval;
   }

   if (jh == num_channels) {		/* bottom (below all channels) */
      thischan = channels + num_channels - 1;
      ly = thischan->yval - BOX_TO_CHAN - via_height;
      if (ly < thisseg->l.y)
	 ly = thisseg->l.y;
   }
   else if (jh != -1) {
      thischan = channels + jh;
      ly = thischan->yval;
   }

   /* If jumpered, and jog is less than a via width, move the y-position of the */
   /* jog to the vertical position of the closest jumper.			*/
   /* (Watch out for conflicts from dogleg route adjustments!)			*/

   if ((thisseg->yjog != -1) && (jh != -1)) {
      if (ABS(thisseg->u.x - thisseg->l.x) < via_width) {
         if (ABS(uy - thisseg->yjog) < ABS(thisseg->yjog - ly))
	    thisseg->yjog = uy;
         else
	    thisseg->yjog = ly;
      }
   }

   /* Set x positions of the lower and upper jumper end contacts */

   if (uy < thisseg->yjog)
      ux = thisseg->l.x;
   else if (uy > thisseg->yjog)
      ux = thisseg->u.x;
   else
      ux = (thisseg->u.x + thisseg->l.x) / 2;

   if (ly < thisseg->yjog)
      lx = thisseg->l.x;
   else if (ly > thisseg->yjog)
      lx = thisseg->u.x;
   else
      lx = (thisseg->u.x + thisseg->l.x) / 2;

   /* Below, each case is done separately.  This can probably be cleaned up	*/
   /* quite a bit by removing redundancy.					*/

   /* Route all metal 1 */

   fprintf(fm, "<< metal1 >>\n");

   if (thisseg->yjog == -1 && jh == -1) {	/* no jog, no jumper */
      outrectv(fm, thisseg->l.x, thisseg->l.y, m1_width, thisseg->u.y);
   }
   else if (thisseg->yjog != -1 && jh == -1) {	/* jogged, no jumper */
      outrectv(fm, thisseg->l.x, thisseg->l.y, m1_width, thisseg->yjog + m1_width);
      outrectv(fm, thisseg->u.x, thisseg->yjog, m1_width, thisseg->u.y);
      if (thisseg->u.x > thisseg->l.x + m1_width)
         outrecth(fm, thisseg->l.x + m1_width, thisseg->yjog, thisseg->u.x, m1_width);
      else if (thisseg->u.x < thisseg->l.x - m1_width)
         outrecth(fm, thisseg->u.x + m1_width, thisseg->yjog, thisseg->l.x, m1_width);
   }
   else if (thisseg->yjog == -1 && jh != -1) {	/* straight jumpered */
      outrectv(fm, thisseg->l.x, thisseg->l.y, m1_width, ly);
      outrectv(fm, thisseg->l.x, uy, m1_width, thisseg->u.y);
   }
   else {				/* jogged and jumpered */
      if (thisseg->yjog < ly) {	/* Case 1: jog is below the jumper */
         outrectv(fm, thisseg->u.x, uy, m1_width, thisseg->u.y);

         outrectv(fm, thisseg->l.x, thisseg->l.y, m1_width, thisseg->yjog);
         outrectv(fm, thisseg->u.x, thisseg->yjog, m1_width, ly);
         if (thisseg->u.x > thisseg->l.x + m1_width)
            outrecth(fm, thisseg->l.x + m1_width, thisseg->yjog, thisseg->u.x, m1_width);
         else if (thisseg->u.x < thisseg->l.x - m1_width)
            outrecth(fm, thisseg->u.x + m1_width, thisseg->yjog, thisseg->l.x, m1_width);
      }
      else if (thisseg->yjog > uy) {	/* Case 2: jog is above the jumper */
         outrectv(fm, thisseg->l.x, thisseg->l.y, m1_width, ly);

         outrectv(fm, thisseg->l.x, uy, m1_width, thisseg->yjog);
         outrectv(fm, thisseg->u.x, thisseg->yjog, m1_width, thisseg->u.y);
         if (thisseg->u.x > thisseg->l.x + m1_width)
            outrecth(fm, thisseg->l.x + m1_width, thisseg->yjog, thisseg->u.x, m1_width);
         else if (thisseg->u.x < thisseg->l.x - m1_width)
            outrecth(fm, thisseg->u.x + m1_width, thisseg->yjog, thisseg->l.x, m1_width);
      }
      else {			/* Case 3: jog is in the middle of the jumper */
         outrectv(fm, thisseg->l.x, thisseg->l.y, m1_width, ly);
         outrectv(fm, thisseg->u.x, uy, m1_width, thisseg->u.y);
      }
   }

   /* Route all metal 2 (only required when route is jumpered) */

   if ((thisseg->yjog == -1 && jh != -1) || (thisseg->yjog != -1 && jh != -1)) {
      fprintf(fm, "<< metal2 >>\n");
   }
   if (thisseg->yjog == -1 && jh != -1) {	/* straight jumpered */
      outrectv(fm, thisseg->l.x, ly, m2_width, uy);
   }
   else if (thisseg->yjog != -1 && jh != -1) {
      if (thisseg->yjog < ly) {	/* Case 1: jog is below the jumper */
         outrectv(fm, thisseg->u.x, ly, m2_width, uy);
      }
      else if (thisseg->yjog > uy) {	/* Case 2: jog is above the jumper */
         outrectv(fm, thisseg->l.x, ly, m2_width, uy);
      }
      else {			/* Case 3: jog is in the middle of the jumper */
         outrectv(fm, thisseg->l.x, ly, m2_width, thisseg->yjog);
         outrectv(fm, thisseg->u.x, thisseg->yjog, m2_width, uy);
         if (thisseg->u.x > thisseg->l.x + m2_width)
            outrecth(fm, thisseg->l.x + m2_width, thisseg->yjog, thisseg->u.x, m2_width);
         else if (thisseg->u.x < thisseg->l.x - m2_width)
            outrecth(fm, thisseg->u.x + m2_width, thisseg->yjog, thisseg->l.x, m2_width);
      }
   }

   /* Default position for an I/O contact is on the channel */

   vx = VIA_X(thisseg->contact.x);
   vy = VIA_Y(thisseg->contact.y);

   /* Place contacts at jumper ends, reposition I/O contact if necessary */

   if ((jl != num_channels) || (jh != -1)) {
      fprintf(fm, "<< via >>\n");
      if (jh != -1) {
         vx = VIA_X(lx);
         vy = (jh == num_channels) ? ly : VIA_Y(ly);
         outrect(fm, vx, vy, via_width, via_height);
      }
      if (jl != num_channels) {
         vx = VIA_X(ux);
         vy = (jl == -1) ? uy : VIA_Y(uy);
         outrect(fm, vx, vy, via_width, via_height);
      }
   }

   /* Add Input/Output labels to m2					 */
   /* ("int" and "ext" names are not Input/Output (xcircuit convention)) */

   if (((networks + thisseg->net)->name != NULL) &&
         (strstr((networks + thisseg->net)->name, "int") == NULL) &&
         (strstr((networks + thisseg->net)->name, "ext") == NULL)) {

      /* If the route is an I/O port but doesn't have a via placed, then  */
      /* generate one to put the label on.  This is only for fet designs, */
      /* where the route may be m1 only.				  */

      if (fet_design && (jh == -1)) {
         fprintf(fm, "<< via >>\n");
	 if (vx < 0) {
	    vx = thisseg->u.x;
	    vy = channels->yval;
            outrect(fm, VIA_X(vx), vy, via_width, via_height);
	 }
	 else
            outrect(fm, vx, vy, via_width, via_height);
      }

      fprintf(fm, "<< labels >>\n");
      fprintf(fm, "rlabel metal2 %d %d %d %d 7 %s\n", vx, vy, vx + m2_width,
            vy + m2_width, (networks + thisseg->net)->name);
      /* fprintf(fm, "port %d e\n", ...); */ /* to be completed. . . */

      /* De-allocate names as we go, so that we don't repeat an I/O label */

      free((networks + thisseg->net)->name);
      (networks + thisseg->net)->name = NULL;
   }
}

/*--------------------------------------------------------------*/
/* Generate a magic layout (.mag) file				*/
/* Return 0 on success, 1 if something could not be properly	*/
/*  routed (e.g., ran of of room for dogleg vertical routes)	*/
/*--------------------------------------------------------------*/

int gen_magic(chstructptr channels, int num_channels, FILE *fm)
{
   int chgt;			/* channel height */
   int uspace, lspace, begin;	// to skip blank space at the beginning
   int i, j, k, l, p1, p2, rowht, ulength, llength, length;
   char this_direc, next_direc;
   int x0, x00, y0, numports, lastlabel;
   cellptr thiscell, nextcell, *source;
   int net, thisnet, psnum = 0, nsnum = 0;
   portinfo *cvals;
   vsave l1route, l2route;
   m1save m1route;
   boxptr thisbox, prevbox, caps, ncap, pcap;
   int status = 0;
   int num_vsegs;
   vsegmentptr vsegs;
   chstructptr thischan;
   int (*vroute_proc)();

   /* Recompute the channel value for each net		*/
   /* (previous values were scrambled by qsort()).	*/

   for (i = 0; i < num_channels; i++) {
      for (j = 0; j < (channels + i)->num_nets; j++) {
	 thisnet = *((channels + i)->nets + j);
	 (networks + thisnet)->channel = i;
      }
   }

   /* determine whether cap cells need to be added to either side	*/
   /* if caps are present, shift other cells to the right to make space	*/
   if (fet_design) {
      ncap = read_box("ncap");
      pcap = read_box("pcap");
      if (ncap == NULL || pcap == NULL || ncap->w != pcap->w)
	 caps = NULL;
      else
	 caps = ncap;
   }
   else
      caps = read_box("cap");
   if (caps != NULL) {
      for (i = 0; i < ROWS; i++) {
	 source = (i ? ucells : lcells);
	 for (j = 0; j < cols; j++) {
	    thiscell = (*(source + j));
	    numports = boxarr[thiscell->boxnum].numports;
	    for (k = 0; k < numports; k++)
	       if (thiscell->pos != NULL)
		  thiscell->pos[k].x += caps->w;
	 }
      }
      for (i = 0; i < netcount; i++) {
	 (networks + i)->begin += caps->w;
	 (networks + i)->end += caps->w;
      }
   }

   /* check for spacer.mag if it hasn't already been read.  set sprnum	*/
   if (!fet_design && !vdd_line && !gnd_line && read_box("spacer") != NULL)
      sprnum = boxcount - 1;
   else if (fet_design) {
      if (read_box("nspacer") != NULL)
	 nsnum = boxcount - 1;
      if (read_box("pspacer") != NULL)
	 psnum = boxcount - 1;
      if ((psnum && !nsnum) || (nsnum && !psnum))
	 fprintf(stderr, "Note: Found \"%s\" but not \"%s\".\n", (nsnum ?
		  "nspacer.mag" : "pspacer.mag"), (nsnum ? "pspacer.mag" :
		  "nspacer.mag"));
   }

   /* Ignore unneeded blank cells at the beginning -- determine offset	*/
   /* of first non-blank cells.						*/
   begin = -(calc_initial_blanks());
   
   /* Determine length of layout without blanks at beginning and end	*/
   uspace = lspace = 0;
   ulength = llength = begin + (caps == NULL ? 0 : caps->w);
   for (i = 0; i < ROWS; i++) {
      source = (i ? ucells : lcells);
      for (j = (cols - 1); ((*(source + j))->order == 0) && (j > 0); j--) {
	 if (i)
	    uspace += boxarr[(*(source + j))->boxnum].w;
	 else
	    lspace += boxarr[(*(source + j))->boxnum].w;
      }
      for (j = 0; j < cols; j++) {
	 if (i)
	    ulength += boxarr[(*(source + j))->boxnum].w;
	 else
	    llength += boxarr[(*(source + j))->boxnum].w;
      }
   }
   length = (ulength - uspace > llength - lspace ? ulength - uspace : llength -
	 lspace);
   
   /* Compute total channel height */

   chgt = (num_channels - 1) * CHAN_SEP + m1_width + 2 * BOX_TO_CHAN;

   /* Begin magic file output */

   if (techname[0] == '\0') {
      fprintf(stderr, "Warning:  No technology read from .mag files.  "
		"Defaults to \"scmos\"\n");
      strcpy(techname, "scmos");
   }
   fprintf(fm, "magic\ntech %s\ntimestamp 0\n", techname);

   y0 = 0;
   for (k = 0; k < ROWS; k++) {
      x0 = begin;
      source = (k == 0 ? lcells : ucells);
      if (fet_design) {
	 if (caps != NULL)
	    caps = (k ? pcap : ncap);
	 sprnum = (k ? psnum : nsnum);
      }
      
      /* place boxes in row k, and log				*/
      /* the y-positions of ports.				*/

      for (i = 0; i < cols; i++) {
	 if (x0 == 0 && caps != NULL) {
	    draw_box(fm, NULL, x0, y0, caps);
	    x0 += caps->w;
	 }
         thiscell = *(source + i);
         thisbox = boxarr + thiscell->boxnum;
         if (!sprnum && thiscell->order == 0) {	/* blank cell */
            if ((x0 >= 0) && (x0 <= length)) {
	       // connect gnd and vdd through spacers
	       fprintf(fm, "<< metal1 >>\n");
	       if (!(fet_design && k != LOWER))
		  outrectv(fm, x0, thisbox->gnd.ll.y + y0, thisbox->w,
			thisbox->gnd.ur.y + y0);
	       if (!(fet_design && k != UPPER))
		  outrectv(fm, x0, thisbox->vdd.ll.y + y0, thisbox->w,
			thisbox->vdd.ur.y + y0);
	    }
	    x0 += thisbox->w;
	    continue;
         }
	 if (fet_design && thiscell->order == 0)
	    thisbox = boxarr + sprnum;
	 if (x0 >= 0 && (x0 + boxarr[thiscell->boxnum].w) <= length)
            draw_box(fm, thiscell, x0, y0, thisbox);

	 // Does not work. . . needs to reflect new port index method
         /* x0 += ((thiscell->overlap) ?
		(thisbox->ports[thisbox->ports[ISOURCE].idx].ll.x - 3) :
		thisbox->w); */

	 x0 += thisbox->w;
	 if (x0 >= length && caps != NULL) {
	    draw_box(fm, NULL, x0, y0, caps);
	    break;
	 }
      }
      if ((x0 < length || !sprnum) && caps != NULL)
	 draw_box(fm, NULL, x0, y0, caps);

      // label layout's GND and Vdd lines
      
/*      fprintf(fm, "<< labels >>\n");
      for (i = 0; i < cols && (*(source + i))->order == 0 ||
	    (*(source + i))->order == gnd_line ||
	    (*(source + i))->order == vdd_line; i++);
      thisbox = boxarr + (*(source + i))->boxnum;
      if (k == LOWER)
	 fprintf(fm, "rlabel metal1 0 %d %d %d 1 GND\n", thisbox->gnd.ll.y + y0,
	       length + (caps == NULL ? 0 : caps->w), thisbox->gnd.ur.y + y0);
      else
	 fprintf(fm, "rlabel metal1 0 %d %d %d 5 VDD\n", thisbox->vdd.ll.y + y0,
	       length + (caps == NULL ? 0 : caps->w), thisbox->vdd.ur.y + y0);
*/
      /* Do direct cell-to-cell routes on row k */

      fprintf(fm, "<< %s >>\n", (fet_design ? "metal1" : "metal2"));
      for (i = 0; i < cols - 1; i++) {
         while ((*(source + i))->order == 0 && i < cols - 1) i++;
         if (i == cols - 1)
	    break;
         thiscell = *(source + i);
         for (j = 1; i + j < cols && (*(source + i + j))->order == 0; j++);
         if (i + j == cols)
	    break;
         nextcell = *(source + i + j);

	 // Draw any ptp connections to gnd/vdd cells

	 if (thiscell->order == vdd_line || thiscell->order == gnd_line ||
	       nextcell->order == vdd_line || nextcell->order == gnd_line)
	 {
	    draw_gv_ptp(fm, thiscell, nextcell);
	    continue;
	 }
         this_direc = (thiscell->orient ? 'w' : 'e');
         next_direc = (nextcell->orient ? 'e' : 'w');
         for (p1 = 0; p1 < boxarr[thiscell->boxnum].numports; p1++)
	    if (check_for_ptp(thiscell, nextcell, this_direc, next_direc,
	   	     p1, &p2, k))
	       point_to_point(fm, thiscell, nextcell, p1, p2);
      }
      
      // Calculate height of bottom row and reset y0 to place upper row boxes

      if (k == LOWER) {
	 rowht = 0;
         for (i = 0; i < cols; i++)
	    if (boxarr[(*(source + i))->boxnum].ht > rowht)
	       rowht = boxarr[(*(source + i))->boxnum].ht;
         y0 += rowht + chgt;
      }
   }

   /* Add y position of each channel to the channel record */

   y0 = rowht + BOX_TO_CHAN;
   for (i = 0; i < num_channels; i++) {
      thischan = channels + i;
      thischan->yval = y0 + CHAN_SEP * (num_channels - i - 1);
   }

   /* Do the channel routes */

   /* Make a list of vertical lines, working from left to right. */
   /* Include direct vertical point-to-point routes but not GND	 */
   /* and Vdd connections (which have been routed, above).	 */

   num_vsegs = 0;
   vsegs = (vsegmentptr)malloc(sizeof(vsegment));

   for (i = 0; i < total_ports; i++) {
      cvals = portorder + i;
      vsegs = (vsegmentptr)realloc(vsegs, (num_vsegs + 1) * sizeof(vsegment));
      if (vroute_prep(cvals, channels, num_channels, vsegs + num_vsegs) >= 0)
         num_vsegs++;
   }
   if (dogleg(vsegs, num_vsegs, channels, num_channels) < 0)
      fprintf(stderr, "Warning:  Cannot resolve all vertical routes.  Hand-editing "
		"required.\n");

   /* Draw the vertical lines */

   vroute_proc = (fet_design) ? vroute : vroute2;

   for (i = 0; i < num_vsegs; i++)
      (*vroute_proc)(fm, vsegs + i, channels, num_channels);

   /* And finally, route the horizontal lines */

   fprintf(fm, "<< metal1 >>\n");
   for (i = 0; i < num_channels; i++) {
      for (j = 0; j < (channels + i)->num_nets; j++) {
	 thisnet = *((channels + i)->nets + j);
	 if (((networks + thisnet)->end - m1_width -
		  (networks + thisnet)->begin) >= PORT_SEP)
	    outrecth(fm, (networks + thisnet)->begin, (channels + i)->yval,
	          (networks + thisnet)->end, m1_width);
      }
   }

   /* end */

   fprintf(fm, "<< end >>\n");
   fflush(fm);

   return status;
}

/*------------------------------------------------------------------*/
